package se.su.dsv.scipro.grading;

import org.glassfish.jersey.jackson.JacksonFeature;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import se.su.dsv.scipro.util.Either;

import javax.inject.Inject;
import javax.inject.Named;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.Entity;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.core.GenericType;
import javax.ws.rs.core.Response;
import java.io.InputStream;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Optional;

public class GradingServiceImpl implements GradingService {

    private static final Logger LOG = LoggerFactory.getLogger(GradingServiceImpl.class);
    private static final GenericType<List<Examination>> EXAMINATION_LIST = new GenericType<List<Examination>>() {};
    private final String url;

    @Inject
    public GradingServiceImpl(@Named("service.grading.url") final String url) {
        this.url = url;
    }

    @Override
    public List<Examination> getExaminations(final String token, final long projectId, final long authorId) {
        final WebTarget client = service();
        final Response response = client
                .path("thesis").path(Long.toString(projectId))
                .path("author").path(Long.toString(authorId))
                .request()
                .header("Authorization", "Bearer " + token)
                .get();
        if (response.getStatus() == Response.Status.OK.getStatusCode()) {
            return response.readEntity(EXAMINATION_LIST);
        }
        else {
            return Collections.emptyList();
        }
    }

    @Override
    public Optional<Result> getResult(final String token, final long projectId, final long authorId, final long examinationId) {
        final WebTarget client = service();
        final Response response = client
                .path("thesis").path(Long.toString(projectId))
                .path("author").path(Long.toString(authorId))
                .path("examination").path(Long.toString(examinationId))
                .request()
                .header("Authorization", "Bearer " + token)
                .get();
        if (response.getStatus() == Response.Status.OK.getStatusCode()) {
            final ResultContainer result = response.readEntity(ResultContainer.class);
            return Optional.of(result.getResult());
        }
        else {
            return Optional.empty();
        }
    }

    @Override
    public Either<ReportGradeError, Void> reportGrade(
            final String token,
            final long projectId,
            final long authorId,
            final int examinationId,
            final String grade,
            final Date reported)
    {
        final Response response = service()
                .path("thesis").path(Long.toString(projectId))
                .path("author").path(Long.toString(authorId))
                .path("examination").path(Long.toString(examinationId))
                .request()
                .header("Authorization", "Bearer " + token)
                .post(Entity.json(new ReportGradeRequest(grade, reported)));
        if (response.getStatus() == Response.Status.OK.getStatusCode()) {
            return Either.right(null);
        }
        else {
            final ErrorContainer errorContainer = response.readEntity(ErrorContainer.class);
            return Either.left(errorContainer.getError());
        }
    }

    @Override
    public Either<UploadProtocolError, Void> sendProtocol(final String token, final long projectId, final long authorId, final int examinationId, final byte[] data, final String contentType) {
        final Response response = service()
                .path("thesis").path(Long.toString(projectId))
                .path("author").path(Long.toString(authorId))
                .path("examination").path(Long.toString(examinationId))
                .path("protocol")
                .request()
                .header("Authorization", "Bearer " + token)
                .post(Entity.entity(data, contentType));
        if (response.getStatus() == Response.Status.OK.getStatusCode()) {
            return Either.right(null);
        }
        else {
            return Either.left(UploadProtocolError.ANY_ERROR);
        }
    }

    @Override
    public Either<UploadTextMatchingError, Void> sendTextMatchingDocument(final String token, final long projectId, final InputStream data, final String contentType) {
        final Response response = service()
                .path("thesis").path(Long.toString(projectId))
                .path("plagiarism")
                .request()
                .header("Authorization", "Bearer " + token)
                .post(Entity.entity(data, contentType));
        if (response.getStatus() == Response.Status.OK.getStatusCode()) {
            return Either.right(null);
        }
        else {
            LOG.info(response.readEntity(String.class));
            return Either.left(UploadTextMatchingError.ANY_ERROR);
        }
    }

    @Override
    public boolean hasTextMatchingDocument(final String token, final long projectId) {
        final Response response = service()
                .path("thesis").path(Long.toString(projectId))
                .path("plagiarism")
                .request()
                .header("Authorization", "Bearer " + token)
                .get();
        return response.getStatus() == Response.Status.OK.getStatusCode();
    }

    private WebTarget service() {
        return ClientBuilder.newBuilder()
                .hostnameVerifier((hostname, sslSession) -> true)
                .register(JacksonFeature.class)
                .build()
                .target(url);
    }
}
