package se.su.dsv.scipro.plagiarism.urkund;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.http.HttpResponse;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.InputStreamEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import se.su.dsv.scipro.file.FileDescription;

import javax.inject.Inject;
import javax.inject.Provider;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.Base64;
import java.util.List;
import java.util.Optional;

/**
 * URKUND allows multiple submissions using the same external id. SciPro is
 * build under the assumption that we will not do this and therefore expects
 * only one {@link SubmissionInfo} when fetching submission status.
 */
class UrkundApiImpl implements UrkundApi {
    private static final int MAX_RECEIVERS_PER_PAGE = 100;

    private final ObjectMapper objectMapper;
    private final Provider<UrkundSettings> urkundSettings;

    @Inject
    UrkundApiImpl(final Provider<UrkundSettings> urkundSettings) {
        this.urkundSettings = urkundSettings;
        objectMapper = new ObjectMapper();
        objectMapper.findAndRegisterModules();
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
    }

    @Override
    public UrkundResponse submitDocument(final String receiver, final FileDescription file, final String submitterEmail, final String subject, final String message, final boolean anonymous) {
        try {
            final HttpPost httpPost = new HttpPost("https://secure.urkund.com/api/rest/submissions/" + receiver + "/" + file.getId());
            httpPost.addHeader("x-urkund-filename", urkundFilename(file));
            httpPost.addHeader("x-urkund-submitter", file.getUploader().getEmailAddress());
            httpPost.addHeader("Authorization", "Basic " + credentials());
            httpPost.addHeader("Accept", "application/json");
            httpPost.setEntity(new InputStreamEntity(file.getData(), file.getSize(), ContentType.create(file.getMimeType())));

            return sendToUrkund(httpPost);
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    private String credentials() {
        final UrkundSettings settings = urkundSettings.get();
        final String credentials = settings.getUsername() + ":" + settings.getPassword();
        return Base64.getMimeEncoder(Integer.MAX_VALUE, new byte[] {'\r', '\n'})
                .encodeToString(credentials.getBytes(StandardCharsets.UTF_8));
    }

    private String urkundFilename(final FileDescription file) {
        final Charset charset = StandardCharsets.UTF_8;
        final byte[] encoded = Base64.getEncoder()
                .encode(file.getName().getBytes(charset));
        return new String(encoded, charset);
    }

    @Override
    public UrkundResponse submissionInfo(final String receiver, final FileDescription file) {
        final String analysisAddress = analysisAddress(receiver).orElse(receiver);
        try {
            final HttpGet httpPost = new HttpGet("https://secure.urkund.com/api/rest/submissions/" + analysisAddress + "/" + file.getId());
            httpPost.addHeader("Authorization", "Basic " + credentials());
            httpPost.addHeader("Accept", "application/json");

            return sendToUrkund(httpPost);
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public Optional<String> analysisAddress(final String email) {
        final HttpGet httpGet = new HttpGet("https://secure.urkund.com/api/receivers?EmailAddress=" + email);
        httpGet.addHeader("Authorization", "Basic " + credentials());
        httpGet.addHeader("Accept", "application/json");
        final List<ReceiverInfo> receiverInfos = sendToUrkund_(httpGet, response -> objectMapper.<List<ReceiverInfo>>readValue(
                response.getEntity().getContent(),
                objectMapper.getTypeFactory().constructCollectionType(List.class, ReceiverInfo.class)));
        return receiverInfos.stream()
                .filter(ri -> ri.getEmailAddress().equals(email))
                .map(ReceiverInfo::getAnalysisAddress)
                .findAny();
    }

    private interface FailingFunction<A, B> {
        B apply(A a) throws Exception;
    }

    private <A> A sendToUrkund_(final HttpUriRequest request, final FailingFunction<HttpResponse, A> f) {
        try (final CloseableHttpClient client = HttpClients.createDefault();
             final CloseableHttpResponse response = client.execute(request))
        {
            return f.apply(response);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    private UrkundResponse sendToUrkund(final HttpUriRequest request) throws IOException {
        return sendToUrkund_(request, response -> {
            switch (response.getStatusLine().getStatusCode()) {
                case 200:
                    final List<SubmissionInfo> submissionInfos = objectMapper.<List<SubmissionInfo>>readValue(response.getEntity().getContent(), objectMapper.getTypeFactory().constructCollectionType(List.class, SubmissionInfo.class));
                    return submissionInfos.get(0);
                case 202:
                    return objectMapper.readValue(response.getEntity().getContent(), SubmissionInfo.class);
                case 400:
                case 404:
                case 415:
                case 413:
                    return objectMapper.readValue(response.getEntity().getContent(), ErrorInfo.class);
                default:
                    final ByteArrayOutputStream bos = new ByteArrayOutputStream((int) response.getEntity().getContentLength());
                    response.getEntity().writeTo(bos);
                    final String body = bos.toString();
                    throw new RuntimeException("Unknown response from URKUND: " + response.getStatusLine().getStatusCode() +": " + body);
            }
        });
    }
}
