package se.su.dsv.scipro.reviewing;

import com.querydsl.core.BooleanBuilder;
import com.querydsl.core.types.Predicate;
import com.querydsl.core.types.dsl.BooleanExpression;
import com.querydsl.core.types.dsl.Expressions;
import com.querydsl.core.types.dsl.StringPath;
import com.querydsl.jpa.JPAExpressions;
import com.querydsl.jpa.impl.JPAQuery;
import org.springframework.data.domain.Pageable;
import se.su.dsv.scipro.finalseminar.QFinalSeminar;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectStatus;
import se.su.dsv.scipro.project.QProject;
import se.su.dsv.scipro.reusable.SciProUtilities;
import se.su.dsv.scipro.system.AbstractServiceImpl;
import se.su.dsv.scipro.system.ProjectType;

import javax.inject.Inject;
import javax.inject.Provider;
import javax.persistence.EntityManager;
import java.util.Collection;
import java.util.Date;

import static com.querydsl.core.types.dsl.Expressions.anyOf;

public class ProjectFinalSeminarStatisticsServiceImpl extends AbstractServiceImpl<Project, Long> implements ProjectFinalSeminarStatisticsService {

    @Inject
    protected ProjectFinalSeminarStatisticsServiceImpl(Provider<EntityManager> em) {
        super(em, Project.class, QProject.project);
    }

    @Override
    public Long countFinalSeminarStatistics(ProjectFinalSeminarStatisticsService.Filter filter) {
        return new JPAQuery<>(em())
                .from(QProject.project, QFinalSeminar.finalSeminar)
                .where(toPredicate(filter))
                .where(QFinalSeminar.finalSeminar.project.eq(QProject.project))
                .fetchCount();
    }

    @Override
    public Long countFinalSeminarStatisticsOkFromReviewer(ProjectFinalSeminarStatisticsService.Filter filter) {
        return new JPAQuery<>(em())
                .from(QProject.project, QFinalSeminar.finalSeminar, QFinalSeminarApproval.finalSeminarApproval)
                .where(toPredicate(filter))
                .where(QFinalSeminar.finalSeminar.project.eq(QProject.project))
                .where(QFinalSeminarApproval.finalSeminarApproval.project.eq(QProject.project),
                        QFinalSeminarApproval.finalSeminarApproval.instanceOf(FinalSeminarApproval.class),
                        QFinalSeminarApproval.finalSeminarApproval.decisions.any().status.eq(Status.APPROVED))
                .fetchCount();
    }

    @Override
    public Iterable<Project> finalSeminarStatisticsNoOkFromReviewer(ProjectFinalSeminarStatisticsService.Filter filter, Pageable pageable) {
        StringPath sort = Expressions.stringPath(QProject.project, pageable.getSort().getProperty());
        return new JPAQuery<>(em())
                .select(QProject.project)
                .from(QProject.project, QFinalSeminar.finalSeminar)
                .where(toPredicate(filter))
                .where(QFinalSeminar.finalSeminar.project.eq(QProject.project))
                .where(noOk())
                .offset(pageable.getOffset())
                .limit(pageable.getLimit())
                .orderBy(pageable.getSort().isAscending() ? sort.asc() : sort.desc())
                .fetch();
    }

    private BooleanExpression noOk() {
        return anyOf(
                QProject.project.notIn(JPAExpressions
                            .select(QFinalSeminarApproval.finalSeminarApproval.project)
                            .from(QFinalSeminarApproval.finalSeminarApproval)
                            .where(QFinalSeminarApproval.finalSeminarApproval.decisions.any().status.eq(Status.APPROVED))),
                JPAExpressions
                        .selectFrom(QFinalSeminarApproval.finalSeminarApproval)
                        .where(QFinalSeminarApproval.finalSeminarApproval.project.eq(QProject.project))
                        .notExists());
    }

    @Override
    public Long countFinalSeminarStatisticsNoOkFromReviewer(ProjectFinalSeminarStatisticsService.Filter filter) {
        return new JPAQuery<>(em())
                .from(QProject.project, QFinalSeminar.finalSeminar)
                .where(toPredicate(filter))
                .where(QFinalSeminar.finalSeminar.project.eq(QProject.project))
                .where(noOk())
                .fetchCount();
    }

    private static Predicate toPredicate(final Filter filter) {
        BooleanBuilder predicate = new BooleanBuilder();
        if (filter.getProjectTypes() != null) {
            predicate.and(levelFilter(filter.getProjectTypes()));
        }
        if (filter.getCreatedAfter() != null) {
            predicate.and(createdAfter(filter.getCreatedAfter()));
        }
        if (filter.getCreatedBefore() != null) {
            predicate.and(createdBefore(filter.getCreatedBefore()));
        }
        if (filter.getStatuses() != null) {
            predicate.and(hasAnyStatus(filter.getStatuses()));
        }
        predicate.and(QFinalSeminar.finalSeminar.dateCreated.before(filter.getFinalSeminarCreatedBefore()));
        predicate.and(QFinalSeminar.finalSeminar.dateCreated.after(filter.getFinalSeminarCreatedAfter()));
        return predicate;
    }

    private static BooleanBuilder hasAnyStatus(Collection<ProjectStatus> statuses) {
        BooleanBuilder ee = new BooleanBuilder();
        if (statuses != null && !statuses.isEmpty()) {
            for (ProjectStatus status : statuses) {
                ee.or(QProject.project.projectStatus.eq(status));
            }
        } else {
            ee.and(QProject.project.projectStatus.isNull());
        }
        return ee;
    }

    private static BooleanBuilder levelFilter(Collection<ProjectType> levels) {
        BooleanBuilder e = new BooleanBuilder();
        if (levels != null && !levels.isEmpty()) {
            for (ProjectType level : levels) {
                e.or(QProject.project.projectType.eq(level));
            }
        } else {
            e.and(QProject.project.projectType.isNull());
        }
        return e;
    }

    private static BooleanExpression createdAfter(Date after) {
        Date start = SciProUtilities.setTimeOfDayTo0000(after);
        return QProject.project.dateCreated.after(start);
    }

    private static BooleanExpression createdBefore(Date before) {
        Date end = SciProUtilities.setTimeOfDayTo2359(before);
        return QProject.project.dateCreated.before(end);
    }
}
