package se.su.dsv.scipro.reviewing;

import com.google.common.eventbus.EventBus;
import com.google.inject.persist.Transactional;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.file.FileDescriptionService;
import se.su.dsv.scipro.file.FileUpload;
import se.su.dsv.scipro.misc.DaysService;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.util.Either;
import se.su.dsv.scipro.util.Option;

import javax.inject.Inject;
import javax.inject.Provider;
import javax.persistence.EntityManager;
import java.util.Date;

class RoughDraftApprovalServiceImpl extends AbstractReviewerApprovalServiceImpl<RoughDraftApproval> implements RoughDraftApprovalService {

    private final DaysService daysService;
    private final ReviewerDeadlineSettingsService reviewerDeadlineSettingsService;

    @Inject
    RoughDraftApprovalServiceImpl(
            final Provider<EntityManager> em,
            final EventBus eventBus,
            final FileDescriptionService fileDescriptionService,
            final DaysService daysService,
            final ReviewerDeadlineSettingsService reviewerDeadlineSettingsService)
    {
        super(em, RoughDraftApproval.class, QRoughDraftApproval.roughDraftApproval, eventBus, fileDescriptionService);
        this.daysService = daysService;
        this.reviewerDeadlineSettingsService = reviewerDeadlineSettingsService;
    }

    @Override
    @Transactional
    public Either<AlreadyRequested, RoughDraftApproval> requestApproval(final Project project, FileUpload thesis, final String comment) {
        final FileDescription storedThesis = fileDescriptionService.storeFile(thesis);

        final Either<AlreadyRequested, RoughDraftApproval> roughDraftApproval = findBy(project)
                .map(earlier -> addNewRequest(comment, storedThesis, earlier))
                .getOrElse(() -> Either.right(newApproval(project, storedThesis, comment)));

        final Either<AlreadyRequested, RoughDraftApproval> persisted =
                roughDraftApproval.map(this::save);
        persisted.foreach(rda -> eventBus.post(new RoughDraftApprovalRequestedEvent(rda)));
        return persisted;
    }

    private RoughDraftApproval newApproval(Project project, FileDescription storedThesis, String comment) {
        Date deadline = getDeadline(new Date());
        if (project.getReviewer() == null) {
            // Ugly hack to stop sending reminders before a reviewer is assigned
            deadline = new Date(1583668380000L);
        }
        return new RoughDraftApproval(project, storedThesis, comment, deadline);
    }

    @Override
    protected Date getDeadline(final Date requested) {
        final ReviewerDeadlineSettings reviewerDeadlineSettings = reviewerDeadlineSettingsService.getInstance();
        return daysService.workDaysAfter(requested, reviewerDeadlineSettings.getRoughDraftApproval());
    }

    @Override
    public Option<RoughDraftApproval> findBy(Project project) {
        return Option.of(findOne(QRoughDraftApproval.roughDraftApproval.project.eq(project)));
    }

    public static void main(String[] args) {
        System.out.println(new Date().getTime());
        System.out.println(new Date(120, 2, 8, 12, 53).getTime());
    }

}
