package se.su.dsv.scipro.system;

import com.google.inject.persist.Transactional;
import com.querydsl.core.types.EntityPath;
import com.querydsl.core.types.Predicate;
import com.querydsl.core.types.dsl.Expressions;
import com.querydsl.core.types.dsl.StringPath;
import com.querydsl.jpa.impl.JPAQuery;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;

import javax.inject.Provider;
import javax.persistence.EntityManager;
import java.io.Serializable;
import java.util.List;

public class AbstractServiceImpl<T, ID extends Serializable> extends AbstractRepository implements GenericService<T, ID> {
    private final Class<T> clazz;
    private final EntityPath<T> path;

	protected AbstractServiceImpl(Provider<EntityManager> em, Class<T> clazz, EntityPath<T> path) {
        super(em);
        this.clazz = clazz;
        this.path = path;
    }

    protected JPAQuery<T> createQuery() {
        return from(path);
    }

    private JPAQuery<T> pagedQuery(Pageable pageable) {
        return applyPageable(createQuery(), pageable);
    }

    private JPAQuery<T> applyPageable(JPAQuery<T> query, Pageable pageable) {
        JPAQuery<T> pagedQuery = paged(query, pageable.getOffset(), pageable.getLimit());
        if (pageable.getSort() == null) {
            return pagedQuery;
        }
        return sorted(pagedQuery, pageable.getSort());
    }

    private JPAQuery<T> paged(JPAQuery<T> query, long offset, long limit) {
        return query.limit(limit).offset(offset);
    }

    protected JPAQuery<T> sorted(JPAQuery<T> query, Sort sort) {
        StringPath property = Expressions.stringPath(path, sort.getProperty());
        return query.orderBy(sort.isAscending() ? property.asc() : property.desc());
    }

    @Override
    public long count(Predicate predicate) {
        return createQuery().where(predicate).fetchCount();
    }

    @Override
    public T findOne(Predicate predicate) {
        return createQuery().where(predicate).fetchOne();
    }

    @Override
    public List<T> findAll(Predicate predicate) {
        return createQuery().where(predicate).fetch();
    }

    @Override
    public List<T> findAll(Predicate predicate, Pageable pageable) {
        return pagedQuery(pageable).where(predicate).fetch();
    }

    @Override
    @Transactional
    public <S extends T> S save(S entity) {
        EntityManager entityManager = em();
        if (entityManager.contains(entity)) {
            return entityManager.merge(entity);
        }
        else {
            entityManager.persist(entity);
            return entity;
        }
    }

    @Override
    public T findOne(ID id) {
		return em().find(clazz, id);
	}

    @Override
    public List<T> findAll() {
        return from(path).fetch();
    }

    @Override
    public List<T> findAll(Pageable pageable) {
        return pagedQuery(pageable).fetch();
    }

    @Override
    public long count() {
		return createQuery().fetchCount();
	}

    @Override
    @Transactional
	public void delete(ID id) {
		delete(em().getReference(clazz, id));
	}

    @Override
    @Transactional
    public void delete(T entity) {
        em().remove(entity);
    }
}
