package se.su.dsv.scipro.util;

import java.util.function.Consumer;
import java.util.function.Function;

/**
 * Represents a disjunction: a result that is either an {@code E} or a {@code A}.
 *
 * An instance of {@code Either<E, A>} is either a {@code Left(E)} or {@code Right(A}}.
 *
 * A common use of a disjunction is to explicitly represent the possibility of
 * failure in a result as opposed to throwing an exception. By convention, the
 * left is used for errors and the right is reserved for successes.
 * For example, a function that attempts to parse an integer from a string may
 * have a return type of {@code Either<NumberFormatException, Integer>}.
 * However, since there is no need to actually throw an exception, the type
 * {@code A} chosen for the "left" could be any type representing an error and
 * has no need to actually extend {@code Exception}.
 */
public abstract class Either<E, A> {
    /**
     * Construct an {@code Either<E, A>} by putting an {@code E} on the left.
     */
    public static <E, A> Either<E, A> left(E error) {
        return new Left<>(error);
    }

    /**
     * Construct an {@code Either<E, A>} by putting an {@code A} on the right.
     */
    public static <E, A> Either<E, A> right(A a) {
        return new Right<>(a);
    }

    /**
     * Catamorphism for {@code Either<E, A>}. Runs the first given function if
     * this is a left, otherwise, the second given function.
     */
    public abstract <X> X fold(Function<E, X> ifError, Function<? super A, X> f);

    /**
     * Runs the given function if this is a right and returns its result.
     * If this is a left it does not call the function and returns the {@code E}
     * on the left.
     */
    public <B> Either<E, B> flatMap(Function<? super A, Either<E, B>> f) {
        return fold(Either::left, f);
    }

    /**
     * Runs the given function if this is a right and returns the result on the
     * right. If this is a left it does not call the function and returns the
     * {@code E} on the left.
     */
    public <B> Either<E, B> map(Function<? super A, B> f) {
        return flatMap(f.andThen(Either::right));
    }

    /**
     * @return {@code true} iff this is a left
     */
    public boolean isLeft() {
        return fold(e -> true, a -> false);
    }

    /**
     * @return {@code true} iff this is a right
     */
    public boolean isRight() {
        return fold(e -> false, a -> true);
    }

    /**
     * If this is a right, run the given side effect.
     *
     * Prefer using other methods if at all possible.
     */
    public void foreach(Consumer<? super A> sideEffect) {
        fold(e -> null, a -> { sideEffect.accept(a); return null; });
    }

    @Override
    public String toString() {
        return fold(
                error -> "Left(" + error.toString() + ")",
                a -> "Right(" + a.toString() + ")"
        );
    }

    private static class Right<E, A> extends Either<E, A> {
        private final A a;

        Right(final A a) {
            this.a = a;
        }

        @Override
        public <X> X fold(final Function<E, X> ifError, final Function<? super A, X> f) {
            return f.apply(a);
        }
    }

    private static class Left<E, A> extends Either<E, A> {
        private final E error;

        public Left(final E error) {
            this.error = error;
        }

        @Override
        public <X> X fold(final Function<E, X> ifError, final Function<? super A, X> f) {
            return ifError.apply(error);
        }
    }
}
