package se.su.dsv.scipro.applicationperiod;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.ajax.markup.html.form.AjaxCheckBox;
import org.apache.wicket.extensions.markup.html.repeater.data.grid.ICellPopulator;
import org.apache.wicket.extensions.markup.html.repeater.data.sort.SortOrder;
import org.apache.wicket.extensions.markup.html.repeater.data.table.IColumn;
import org.apache.wicket.extensions.markup.html.repeater.data.table.PropertyColumn;
import org.apache.wicket.extensions.markup.html.repeater.data.table.export.AbstractExportableColumn;
import org.apache.wicket.extensions.model.AbstractCheckBoxModel;
import org.apache.wicket.feedback.FencedFeedbackPanel;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.form.ChoiceRenderer;
import org.apache.wicket.markup.html.form.EnumChoiceRenderer;
import org.apache.wicket.markup.html.form.Form;
import org.apache.wicket.markup.html.link.Link;
import org.apache.wicket.markup.html.panel.FeedbackPanel;
import org.apache.wicket.markup.repeater.Item;
import org.apache.wicket.model.*;
import se.su.dsv.scipro.admin.pages.AdminMailPage;
import se.su.dsv.scipro.components.*;
import se.su.dsv.scipro.components.menuhighlighting.MenuHighlightAdminApplicationPeriod;
import se.su.dsv.scipro.data.DetachableServiceModel;
import se.su.dsv.scipro.dataproviders.FilteredDataProvider;
import se.su.dsv.scipro.datatables.target.EditTargetPanel;
import se.su.dsv.scipro.match.*;
import se.su.dsv.scipro.springdata.services.SupervisorService;
import se.su.dsv.scipro.springdata.services.SupervisorService.FilterParams;
import se.su.dsv.scipro.springdata.services.UnitService;
import se.su.dsv.scipro.system.*;

import javax.inject.Inject;
import java.util.*;

import static se.su.dsv.scipro.data.enums.DateStyle.DATETIME;

public class AdminEditTargetsPage extends AbstractAdminMatchPage implements MenuHighlightAdminApplicationPeriod {

    private static final boolean EDITABLE = true;
    private static final boolean NOT_EDITABLE = false;
    static final String EMPLOYEE_LIST = "employeeList";
    private static final String FILTER_FORM = "filterForm";
    static final String HEADER = "header";
    static final String FEEDBACK = "feedback";

    @Inject
    private SupervisorService supervisorService;
    @Inject
    private ApplicationPeriodService applicationPeriodService;
    @Inject
    private IdeaService ideaService;
    @Inject
    private UnitService unitService;
    @Inject
    private ResearchAreaService researchAreaService;
    @Inject
    private ProjectTypeService projectTypeService;
    @Inject
    private TargetService targetService;

    private IModel<ApplicationPeriod> current;
    private ExportableDataPanel<User,String> table;
    private FeedbackPanel feedback;
    private FilteredDataProvider<User, FilterParams> provider;
    private Link<?> sendMailLink;

    public AdminEditTargetsPage(IModel<ApplicationPeriod> model) {

        this.current = model;
        addFeedback();
        addApplicationPeriodInfo();

        IModel<FilterParams> filter = prepareFilter();
        setUpDataTable(filter);

        add(new FilterForm(FILTER_FORM, filter));

        sendMailLink = new Link("sendMail"){
            @Override
            protected void onConfigure() {
                setVisibilityAllowed(provider.size() > 0);
            }
            @Override
            public void onClick() {
                Set<User> listOfUsers = new HashSet<>();
                Iterator<?> user = provider.iterator(0, provider.size());
                while (user.hasNext()) {
                    listOfUsers.add((User)user.next());
                }
                setResponsePage(new AdminMailPage(listOfUsers));
            }
        };
        sendMailLink.setOutputMarkupPlaceholderTag(true);
        add(sendMailLink);
    }

    private void addFeedback() {
        feedback = new FencedFeedbackPanel(FEEDBACK, this);
        feedback.setOutputMarkupId(true);
        add(feedback);
    }

    private void addApplicationPeriodInfo() {
        add(new Label(HEADER, new StringResourceModel("header", current)));
        add(new DateLabel("startDate", new PropertyModel<>(current, "startDate"), DATETIME));
        add(new DateLabel("endDate", new PropertyModel<>(current, "endDate"), DATETIME));
        add(new DateLabel("courseStartDate", new PropertyModel<>(current, "courseStartDate"), DATETIME));
    }

    private void setUpDataTable(IModel<FilterParams> filter) {
        provider = new FilteredDataProvider<>(supervisorService, filter);
        provider.setSort("firstName", SortOrder.ASCENDING);
        table = new ExportableDataPanel<>(EMPLOYEE_LIST, columns(), provider);
        table.setOutputMarkupId(true);
        add(table);
    }

    private Model<FilterParams> prepareFilter() {
        FilterParams filter = new FilterParams();
        filter.setLanguages(Arrays.asList(Language.values()));
        return Model.of(filter);
    }

    private List<? extends IColumn<User, String>> columns() {
        List<IColumn<User, String>> list = new ArrayList<>();
        list.add(new PropertyColumn<>(new StringResourceModel("name", this), "firstName", "fullName"));
        list.add(new PropertyColumn<>(new StringResourceModel("unit", this), "unit.title"));
        list.add(new PropertyColumn<>(new StringResourceModel("researchAreas", this), "researchAreaNames"));
        addColumnsForProjectTypes(list, current, EDITABLE);
        for (ApplicationPeriod applicationPeriod : applicationPeriodService.getPreviousPeriods(current.getObject(), 2)) {
            addColumnsForProjectTypes(list, new DetachableServiceModel<>(applicationPeriodService, applicationPeriod), NOT_EDITABLE);
        }
        return list;
    }

    private void addColumnsForProjectTypes(List<IColumn<User, String>> list, IModel<ApplicationPeriod> applicationPeriod, boolean isEditable) {
        for (final ProjectType projectType : applicationPeriod.getObject().getProjectTypes()) {
            list.add(matchedColumn(new DetachableServiceModel<>(projectTypeService, projectType), applicationPeriod));
            list.add(editTargetColumn(new DetachableServiceModel<>(projectTypeService, projectType), applicationPeriod, isEditable));
        }
    }

    private IColumn<User, String> editTargetColumn(final IModel<ProjectType> projectType, final IModel<ApplicationPeriod> appPeriod, final boolean isEditable) {
        return new AbstractExportableColumn<User, String>(new StringResourceModel("targetCurrent", appPeriod).setParameters(projectType.getObject().getName()).wrapOnAssignment(this)) {
            @Override
            public void populateItem(Item<ICellPopulator<User>> item, String id, IModel<User> employee) {
                item.add(new EditTargetPanel(id, appPeriod, employee, projectType, isEditable) {
                    @Override
                    public void update(AjaxRequestTarget target, User user) {
                        success(getString("saved", Model.of(user)));
                        target.add(feedback);
                        target.add(table);
                    }
                });
            }
            @Override
            public IModel<Integer> getDataModel(IModel<User> rowModel) {
                return Model.of(targetService.findOne(appPeriod.getObject(), rowModel.getObject(), projectType.getObject()).getTarget());
            }
        };
    }

    private IColumn<User, String> matchedColumn(final DetachableServiceModel<ProjectType> projectType, final IModel<ApplicationPeriod> appPeriod) {
        return new AbstractExportableColumn<User, String>(new StringResourceModel("matchedCurrent", appPeriod).setParameters(projectType.getObject().getName()).wrapOnAssignment(this)) {
            @Override
            public void populateItem(Item<ICellPopulator<User>> item, String id, IModel<User> employee) {
                item.add(new Label(id, Long.toString(ideaService.countMatched(appPeriod.getObject(), employee.getObject(), projectType.getObject()))));
            }

            @Override
            public IModel<Long> getDataModel(IModel<User> rowModel) {
                Long matched = ideaService.countMatched(appPeriod.getObject(), rowModel.getObject(), projectType.getObject());
                return Model.of(matched);
            }
        };
    }

    private class FilterForm extends Form<FilterParams> {
        public FilterForm(String id, final IModel<FilterParams> model) {
            super(id, new CompoundPropertyModel<>(model));
            add(new AjaxCheckBoxMultipleChoice<Language>("languages", Arrays.asList(Language.values()), new EnumChoiceRenderer<>(this)) {
                @Override
                public void onUpdate(AjaxRequestTarget target) {
                    target.add(table, sendMailLink);
                }
            });
            add(new AjaxDropDownChoice<Unit>("unit", unitService.findAll(), new ChoiceRenderer<>("title", "id")) {
                @Override
                public void onNewSelection(AjaxRequestTarget target, Unit objectSelected) {
                    target.add(table, sendMailLink);
                }
            }.setNullValid(true));
            add(new AjaxDropDownChoice<ResearchArea>("researchArea", researchAreaService.findNotDeleted(), new ChoiceRenderer<>("title", "id")) {
                @Override
                public void onNewSelection(AjaxRequestTarget target, ResearchArea objectSelected) {
                    target.add(table, sendMailLink);
                }
            }.setNullValid(true));
            add(new EmployeeAutoComplete("supervisor") {
                @Override
                protected void action(AjaxRequestTarget target, User object) {
                    target.add(table, sendMailLink);
                }
            });
            add(new AjaxCheckBox("hasTargetsInApplicationPeriod", new TargetCheckBoxModel(model)) {
                @Override
                protected void onUpdate(AjaxRequestTarget target) {
                    target.add(table, sendMailLink);
                }
            });
        }
    }

    private class TargetCheckBoxModel extends AbstractCheckBoxModel {
        private IModel<FilterParams> filter;

        TargetCheckBoxModel(IModel<FilterParams> filter) {
            this.filter = filter;
        }

        @Override
        public boolean isSelected() {
            return filter.getObject().getHasTargetsInApplicationPeriod() != null;
        }

        @Override
        public void select() {
            filter.getObject().setHasTargetsInApplicationPeriod(current.getObject());
        }

        @Override
        public void unselect() {
            filter.getObject().setHasTargetsInApplicationPeriod(null);
        }
    }
}
