package se.su.dsv.scipro.datatables.project;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.ajax.form.AjaxFormComponentUpdatingBehavior;
import org.apache.wicket.ajax.markup.html.form.AjaxCheckBox;
import org.apache.wicket.markup.html.form.ChoiceRenderer;
import org.apache.wicket.markup.html.form.EnumChoiceRenderer;
import org.apache.wicket.markup.html.form.TextField;
import org.apache.wicket.markup.html.panel.Panel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.Model;
import org.apache.wicket.model.PropertyModel;
import se.su.dsv.scipro.components.AjaxCheckBoxMultipleChoice;
import se.su.dsv.scipro.components.AuthorAutoComplete;
import se.su.dsv.scipro.components.EmployeeAutoComplete;
import se.su.dsv.scipro.date.DeactivatableDatePickerPanel;
import se.su.dsv.scipro.project.ProjectService;
import se.su.dsv.scipro.project.ProjectStatus;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.ProjectTypeService;
import se.su.dsv.scipro.system.User;

import javax.inject.Inject;

import static java.util.Arrays.asList;

public abstract class ProjectFilterPanel extends Panel {

    private static final int MIN_TITLE_LENGTH = 3;

    @Inject
    private ProjectTypeService projectTypeService;

    public ProjectFilterPanel(final String id, final IModel<ProjectService.Filter> model) {
        super(id, model);

        add(new AjaxCheckBox("noReviewer", new PropertyModel<>(model, "noReviewer")) {
            @Override
            protected void onUpdate(final AjaxRequestTarget target) {
                filterUpdated(target);
            }
        });

        addDateFilter(model);
        addRedStateFilter(model);
        addSupervisorFilter(model);
        addProjectTitleFilter(model);
        addAuthorFilter(model);
        addProjectTypeFilter(model);
        addProjectStatusFilter(model);
    }

    protected abstract void filterUpdated(AjaxRequestTarget target);

    private void addDateFilter(final IModel<ProjectService.Filter> filter) {
        add(new DeactivatableDatePickerPanel("dateFilter",
                new PropertyModel<>(filter, "createdAfter"),
                new PropertyModel<>(filter, "createdBefore")) {
            @Override
            protected void datesChanged(AjaxRequestTarget target) {
                filterUpdated(target);
            }
        });
    }

    private void addRedStateFilter(final IModel<ProjectService.Filter> filter) {
        TextField<Integer> redStateField = new TextField<>("redStateFilter", new PropertyModel<Integer>(filter, "redStateTime"), Integer.class);
        redStateField.add(new AjaxFormComponentUpdatingBehavior("keyup") {
            @Override
            protected void onUpdate(AjaxRequestTarget target) {
                filterUpdated(target);
            }
        });
        add(redStateField);
    }

    private void addSupervisorFilter(final IModel<ProjectService.Filter> filter) {
        add(new EmployeeAutoComplete("supervisorFilter", new PropertyModel<>(filter, "supervisor")) {
            @Override
            protected void action(AjaxRequestTarget target, User newSelection) {
                filterUpdated(target);
            }
        });
    }

    private void addProjectTitleFilter(final IModel<ProjectService.Filter> model) {
        TextField<String> titleField = new TextField<>("titleFilter", new Model<>());
        titleField.add(new AjaxFormComponentUpdatingBehavior("keyup") {
            @Override
            protected void onUpdate(AjaxRequestTarget target) {
                final String titleString = titleField.getModelObject();
                final ProjectService.Filter filter = model.getObject();
                if (titleString != null && titleString.length() >= MIN_TITLE_LENGTH) {
                    filter.setTitleContains(titleString);
                } else {
                    filter.setTitleContains(null);
                }
                filterUpdated(target);
            }
        });
        add(titleField);
    }

    private void addAuthorFilter(final IModel<ProjectService.Filter> filter) {
        add(new AuthorAutoComplete("authorFilter", new PropertyModel<>(filter, "authorUser")) {
            @Override
            protected void action(AjaxRequestTarget target, User newSelection) {
                filterUpdated(target);
            }
        });
    }

    private void addProjectTypeFilter(final IModel<ProjectService.Filter> filter) {
        add(new AjaxCheckBoxMultipleChoice<ProjectType>("projectTypeFilter",
                new PropertyModel<>(filter, "projectTypes"),
                projectTypeService.findAllActive(),
                new ChoiceRenderer<>("name", "id")) {
            @Override
            public void onUpdate(AjaxRequestTarget target) {
                filterUpdated(target);
            }
        });
    }

    private void addProjectStatusFilter(final IModel<ProjectService.Filter> filter) {
        add(new AjaxCheckBoxMultipleChoice<ProjectStatus>("statusFilter",
                new PropertyModel<>(filter, "statuses"),
                asList(ProjectStatus.values()),
                new EnumChoiceRenderer<>(ProjectFilterPanel.this)) {
            @Override
            public void onUpdate(AjaxRequestTarget target) {
                filterUpdated(target);
            }
        });
    }
}
