package se.su.dsv.scipro.grading;

import com.google.common.collect.Sets;
import lombok.SneakyThrows;
import net.sf.jasperreports.engine.JRException;
import org.apache.wicket.markup.html.WebMarkupContainer;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.link.BookmarkablePageLink;
import org.apache.wicket.markup.html.list.ListItem;
import org.apache.wicket.markup.html.list.ListView;
import org.apache.wicket.markup.html.panel.FeedbackPanel;
import org.apache.wicket.markup.html.panel.Fragment;
import org.apache.wicket.model.*;
import org.apache.wicket.request.flow.RedirectToUrlException;
import org.apache.wicket.request.mapper.parameter.PageParameters;
import org.apache.wicket.util.lang.Classes;
import se.su.dsv.scipro.components.ListAdapterModel;
import se.su.dsv.scipro.components.OppositeVisibility;
import se.su.dsv.scipro.daisyExternal.http.DaisyAPI;
import se.su.dsv.scipro.data.FilteredListModel;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.file.FileDescriptionService;
import se.su.dsv.scipro.finalthesis.FinalThesis;
import se.su.dsv.scipro.finalthesis.FinalThesisService;
import se.su.dsv.scipro.io.dto.ThesisPublication;
import se.su.dsv.scipro.oauth.OAuth;
import se.su.dsv.scipro.oauth.OAuthService;
import se.su.dsv.scipro.profile.UserLabel;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.report.*;
import se.su.dsv.scipro.supervisor.pages.AbstractSupervisorProjectDetailsPage;
import se.su.dsv.scipro.system.Language;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.util.Either;

import javax.inject.Inject;
import java.io.IOException;
import java.time.Instant;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

public class SupervisorGradingPage extends AbstractSupervisorProjectDetailsPage {

    @Inject
    private OAuthService oAuthService;
    @Inject
    private GradingService gradingService;
    @Inject
    private GradingReportService gradingReportService;
    @Inject
    private FinalThesisService finalThesisService;
    @Inject
    private DaisyAPI daisyAPI;
    @Inject
    private GradeCalculatorService gradeCalculatorService;
    @Inject
    private FileDescriptionService fileDescriptionService;

    private final IModel<FinalThesis> finalThesisModel;
    private IModel<Boolean> hasFinalThesis;
    private IModel<Boolean> hasTextMatchingDocument;

    public SupervisorGradingPage(final PageParameters pp) {
        super(pp);

        add(new FeedbackPanel("feedback"));

        finalThesisModel = new LoadableDetachableModel<FinalThesis>() {
            @Override
            protected FinalThesis load() {
                return finalThesisService.findByProject(projectModel.getObject());
            }
        };

        hasFinalThesis = new AbstractReadOnlyModel<Boolean>() {
            @Override
            public Boolean getObject() {
                final FinalThesis finalThesis = finalThesisModel.getObject();
                return finalThesis != null && finalThesis.getStatus() == FinalThesis.Status.APPROVED;
            }
        };
        hasTextMatchingDocument = new AbstractReadOnlyModel<Boolean>() {
            @Override
            public Boolean getObject() {
                final FinalThesis finalThesis = finalThesisModel.getObject();
                return finalThesis != null && finalThesis.getTextMatchingDocument() != null;
            }
        };

        final IModel<List<User>> authors = new ListAdapterModel<>(new PropertyModel<>(projectModel, "projectParticipants"));
        add(new ListView<User>("authorsNonGraded", authors) {
            {
                setReuseItems(true);
            }

            @Override
            protected void populateItem(final ListItem<User> item) {
                item.add(new UserLabel("authorName", item.getModel()));
                final IModel<List<Examination>> nonGradedExaminations = getSpecificExaminations(item.getModel(), false);
                item.add(new NonGradedPanel("author", item.getModel(), nonGradedExaminations));
            }
        });

        add(new ListView<User>("authorsGraded", authors) {
            {
                setReuseItems(true);
            }

            @Override
            protected void populateItem(final ListItem<User> item) {
                item.add(new UserLabel("authorName", item.getModel()));
                final IModel<SupervisorGradingReport> gradingReportModel = new LoadableDetachableModel<SupervisorGradingReport>() {
                    @Override
                    protected SupervisorGradingReport load() {
                        return gradingReportService.getSupervisorGradingReport(projectModel.getObject(), item.getModelObject());
                    }
                };
                final IModel<Boolean> hasFinishedGradingReport = new AbstractReadOnlyModel<Boolean>() {
                    @Override
                    public Boolean getObject() {
                        return gradingReportModel.getObject().getState() == GradingReport.State.FINALIZED;
                    }

                    @Override
                    public void detach() {
                        gradingReportModel.detach();
                    }
                };
                final IModel<List<Examination>> gradedExaminations = getSpecificExaminations(item.getModel(), true);
                final ProgressPanel progress = new ProgressPanel("progress", hasFinishedGradingReport);
                item.add(progress);
                final GradedPanel gradedPanel = new GradedPanel("author", item.getModel(), gradedExaminations);
                gradedPanel.add(new OppositeVisibility(progress));
                item.add(gradedPanel);
                item.add(new Label("gradeSuggestion", new AbstractReadOnlyModel<GradingReport.Grade>() {
                    @Override
                    public GradingReport.Grade getObject() {
                        final GradeCalculator gradeCalculator = getGradeCalculator(projectModel.getObject());
                        return gradingReportModel.getObject().getGrade(gradeCalculator);
                    }
                }));
            }
        });
    }

    @Override
    public void detachModels() {
        super.detachModels();
        finalThesisModel.detach();
        hasFinalThesis.detach();
        hasTextMatchingDocument.detach();
    }

    private IModel<List<Examination>> getSpecificExaminations(final IModel<User> model, final boolean graded) {
        final IModel<List<Examination>> loadableDetachableModel = getExaminations(model);
        return new FilteredListModel<Examination>(loadableDetachableModel) {
            @Override
            protected boolean accept(final Examination examination) {
                return isGraded(examination) == graded;
            }
        };
    }

    private IModel<List<Examination>> getExaminations(final IModel<User> model) {
        return new LoadableDetachableModel<List<Examination>>() {
                @Override
                protected List<Examination> load() {
                    return gradingService.getExaminations(
                            getSession().getMetaData(OAuth.TOKEN),
                            projectModel.getObject().getIdentifier(),
                            model.getObject().getIdentifier());
                }
            };
    }

    @Override
    protected void onConfigure() {
        super.onConfigure();
        final Instant expiration = getSession().getMetaData(OAuth.EXPIRATION);
        final boolean validToken = expiration != null && expiration.isAfter(Instant.now());
        if (!validToken) {
            final PageParameters pp = SupervisorGradingPage.getPageParameters(projectModel.getObject());
            final String returnUrl = "/" + getRequestCycle().mapUrlFor(SupervisorGradingPage.class, pp).canonical().toString();
            getSession().setMetaData(OAuth.RETURN_URL, returnUrl);
            final String href = oAuthService.authorizeUrl(
                    null,
                    Sets.newHashSet("grade:read", "grade:write"));
            throw new RedirectToUrlException(href);
        }
    }

    private class NonGradedPanel extends AbstractExaminationsPanel {
        public NonGradedPanel(
                final String id,
                final IModel<User> userModel,
                final IModel<List<Examination>> nonGradedExaminations)
        {
            super(id, projectModel, userModel, nonGradedExaminations);
        }

        @Override
        protected void reportGrade(final Project project, final User author, final Examination examination, final String letter, final Date reported) {
            final Either<ReportGradeError, Void> result = gradingService.reportGrade(
                    getSession().getMetaData(OAuth.TOKEN),
                    project.getIdentifier(),
                    author.getIdentifier(),
                    examination.getId(),
                    letter,
                    reported);
            final String feedback = result.fold(
                    error -> getString(Classes.simpleName(error.getDeclaringClass()) + '.' + error.name()),
                    success -> "Grade " + letter + " reported for " + author.getFullName());
            if (result.isRight()) {
                success(feedback);
            }
            else {
                error(feedback);
            }
        }
    }

    private class GradedPanel extends NonGradedPanel {
        public GradedPanel(
                final String id,
                final IModel<User> userModel,
                final IModel<List<Examination>> gradedExaminations) {
            super(id, userModel, gradedExaminations);
        }

        @Override
        protected void reportGrade(final Project project, final User user, final Examination examination, final String letter, final Date reported) {
            final String token = getSession().getMetaData(OAuth.TOKEN);

            if (!sendPublicationToDaisy(project)) {
                error(getString("publication.could.not.be.sent.to.daisy"));
                return;
            }
            else {
                success(getString("publication.sent.to.daisy"));
            }

            if (!sendTextMatchingDocument(project, token)) {
                error(getString("text.matching.could.not.be.sent.to.daisy"));
                return;
            }
            else {
                success(getString("text.matching.sent.to.daisy"));
            }

            super.reportGrade(project, user, examination, letter, reported);
            if (hasErrorMessage()) {
                return;
            }

            if (!sendProtocol(project, user, examination, token)) {
                error(getString("protocol.could.not.be.sent.to.daisy"));
            }
            else {
                success(getString("protocol.sent.to.daisy"));
            }
        }

        private boolean sendPublicationToDaisy(final Project project) {
            final ThesisPublication old = daisyAPI
                    .getPublication(projectModel.getObject().getIdentifier())
                    .getOrElse(() -> null);
            if (old == null || old.getFile() == null) {
                final ThesisPublication thesisPublication = createThesisPublication();
                final boolean sentPublication = daisyAPI.sendPublication(project.getIdentifier(), thesisPublication);

                if (sentPublication) {
                    final FinalThesis finalThesis = finalThesisModel.getObject();
                    final FileDescription finalThesisDocument = finalThesis.getFileDescription();

                    boolean sentPublicationFile;
                    try {
                        sentPublicationFile = daisyAPI.sendPublicationFile(
                                project.getIdentifier(),
                                finalThesisDocument.getName(),
                                finalThesisDocument.getSize(),
                                fileDescriptionService.getData(finalThesisDocument));
                    } catch (IOException e) {
                        sentPublicationFile = false;
                    }
                    if (!sentPublicationFile) {
                        return false;
                    }
                }
                else {
                    return false;
                }
            }
            return true;
        }

        private ThesisPublication createThesisPublication() {
            final FinalThesis finalThesis = finalThesisModel.getObject();
            final ThesisPublication thesisPublication = new ThesisPublication();
            final Calendar calendar = Calendar.getInstance();
            calendar.setTime(finalThesis.getFileDescription().getDateCreated());
            thesisPublication.setDate(calendar);
            thesisPublication.setLanguage(projectModel.getObject().getLanguage() == Language.SWEDISH ? "sv" : "en");
            return thesisPublication;
        }

        private boolean sendTextMatchingDocument(final Project project, final String token) {
            final boolean hasTextMatchingDocument =
                    gradingService.hasTextMatchingDocument(
                            getSession().getMetaData(OAuth.TOKEN),
                            projectModel.getObject().getIdentifier());
            if (!hasTextMatchingDocument) {
                try {
                    final FinalThesis finalThesis = finalThesisModel.getObject();
                    final Either<UploadTextMatchingError, Void> sendTextMatchingDocument
                            = gradingService.sendTextMatchingDocument(
                                    token,
                                    project.getIdentifier(),
                                    fileDescriptionService.getData(finalThesis.getTextMatchingDocument()),
                                    finalThesis.getTextMatchingDocument().getMimeType());
                    SupervisorGradingPage.this.hasTextMatchingDocument.detach();
                    return sendTextMatchingDocument.isRight();
                } catch (IOException e) {
                    return false;
                }
            }
            return true;
        }

        private boolean sendProtocol(final Project project, final User user, final Examination examination, final String token) {
            final SupervisorGradingReport supervisorGradingReport =
                    gradingReportService.getSupervisorGradingReport(project, user);
            final GradeCalculator gradeCalculator = getGradeCalculator(project);
            final DataExposedGradingReportPdf pdf =
                    new DataExposedGradingReportPdf(
                            supervisorGradingReport,
                            gradeCalculator);
            final Either<UploadProtocolError, Void> sendProtocol = gradingService.sendProtocol(
                    token,
                    project.getIdentifier(),
                    getModelObject().getIdentifier(),
                    examination.getId(),
                    pdf.getData(),
                    pdf.getContentType());
            return sendProtocol.isRight();
        }

        private class DataExposedGradingReportPdf extends GradingReportPdfResource {
            private DataExposedGradingReportPdf(final SupervisorGradingReport gradingReport, final GradeCalculator gradeCalculator) {
                super(new SupervisorGradingReportPdf(Model.of(gradingReport), gradeCalculator));
            }

            @SneakyThrows(JRException.class)
            public byte[] getData() {
                return getExporterData(newJasperPrint(), newExporter());
            }
        }
    }

    private class ProgressPanel extends Fragment {
        private final IModel<Boolean> hasFinishedGradingReport;

        public ProgressPanel(final String id, final IModel<Boolean> hasFinishedGradingReport) {
            super(id, "progress", SupervisorGradingPage.this, hasFinishedGradingReport);
            this.hasFinishedGradingReport = hasFinishedGradingReport;
            add(new RedGreenLabel(
                    "gradingReport",
                    hasFinishedGradingReport,
                    "grading.report.not.done",
                    "grading.report.finished"));
            add(new RedGreenLabel(
                    "finalThesis",
                    hasFinalThesis,
                    "approve.final.thesis",
                    "final.thesis.approved"));
            add(new RedGreenLabel(
                    "textMatching",
                    hasTextMatchingDocument,
                    "upload.text.matching.document",
                    "text.matching.document.uploaded"));
            add(new BookmarkablePageLink<Void>(
                    "finish",
                    SupervisorGradingReportPage.class,
                    SupervisorGradingReportPage.getPageParameters(projectModel.getObject())));
        }

        @Override
        protected void onConfigure() {
            setVisibilityAllowed(!(hasFinishedGradingReport.getObject() && hasFinalThesis.getObject() && hasTextMatchingDocument.getObject()));
        }
    }

    private class RedGreenLabel extends Fragment {
        public RedGreenLabel(final String id, final IModel<Boolean> isGreen, final String redKey, final String greenKey) {
            super(id, "redgreen", SupervisorGradingPage.this);

            queue(new WebMarkupContainer("redContainer") {
                @Override
                protected void onConfigure() {
                    setVisibilityAllowed(!isGreen.getObject());
                }
            });
            queue(new Label("red", new ResourceModel(redKey)) {
                @Override
                protected void onConfigure() {
                    setVisibilityAllowed(!isGreen.getObject());
                }
            });
            queue(new WebMarkupContainer("greenContainer") {
                @Override
                protected void onConfigure() {
                    setVisibilityAllowed(isGreen.getObject());
                }
            });
            queue(new Label("green", new ResourceModel(greenKey)));
        }
    }

    private boolean isGraded(final Examination examination) {
        final long passingGrades = examination.getGrades()
                .stream()
                .map(Grade::getType)
                .filter(Grade.Type.PASSING::equals)
                .count();
        return passingGrades > 1;
    }

    private GradeCalculator getGradeCalculator(final Project project) {
        return gradeCalculatorService.getSupervisorCalculator(project);
    }
}
