package se.su.dsv.scipro.match;

import org.apache.wicket.Component;
import org.apache.wicket.MetaDataKey;
import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.extensions.markup.html.repeater.data.grid.ICellPopulator;
import org.apache.wicket.extensions.markup.html.repeater.data.table.IColumn;
import org.apache.wicket.extensions.markup.html.repeater.data.table.PropertyColumn;
import org.apache.wicket.extensions.markup.html.repeater.data.table.export.AbstractExportableColumn;
import org.apache.wicket.extensions.markup.html.repeater.util.SortableDataProvider;
import org.apache.wicket.markup.html.panel.FeedbackPanel;
import org.apache.wicket.markup.html.panel.Panel;
import org.apache.wicket.markup.repeater.Item;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.Model;
import org.apache.wicket.model.PropertyModel;
import se.su.dsv.scipro.components.*;
import se.su.dsv.scipro.dataproviders.FilteredDataProvider;
import se.su.dsv.scipro.profile.StudentProfileProgramPanel;
import se.su.dsv.scipro.profile.UserLinkPanel;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.ProjectModule;
import se.su.dsv.scipro.system.ProjectTypeService;
import se.su.dsv.scipro.system.User;

import javax.inject.Inject;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import static java.util.Arrays.asList;

public abstract class AbstractAdminIdeaPanel extends Panel {

    public static final MetaDataKey<Long> PRESELECTED_APPLICATION_PERIOD = new MetaDataKey<Long>() {};

    public static final String DIALOG = "dialog";
    public static final String DATA_TABLE = "dataTable";
    public static final String HEADER_DATE_CREATED = "Created";
    public static final String HEADER_PROJECT_TYPE = "Type";
    public static final String HEADER_TITLE = "Title";
    public static final String HEADER_STATUS = "Status";
    public static final String HEADER_SUPERVISOR = "Supervisor";
    public static final String HEADER_AUTHORS = "Authors";
    public static final String TITLE = "title";
    public static final String FEEDBACK = "feedback";
    protected final FeedbackPanel feedback;

    @Inject
    private IdeaService ideaService;
    @Inject
    private TargetService targetService;
    @Inject
    private ApplicationPeriodService applicationPeriodService;
    @Inject
    private ProjectTypeService projectTypeService;

    private ModalWindowPlus modalWindowPlus;
    protected ExportableDataPanel<Idea, String> dataTable;
    private final IdeaService.Filter filter = new IdeaService.Filter();

    public AbstractAdminIdeaPanel(String id) {
        super(id);
        feedback = new FeedbackPanel(FEEDBACK);
        feedback.setOutputMarkupId(true);
        add(feedback);
        addDialog();
        addFiltering();
        addDataTable();
    }

    public IdeaService.Filter getFilter() {
        return filter;
    }

    private void addDialog() {
        modalWindowPlus = new ModalWindowPlus(DIALOG);
        add(modalWindowPlus);
    }

    private void addDataTable() {
        SortableDataProvider<Idea, String> provider = new FilteredDataProvider<>(ideaService, Model.of(filter));
        dataTable = new ExportableDataPanel<>(DATA_TABLE, createColumns(), provider);
        dataTable.setOutputMarkupId(true);
        add(dataTable);
    }

    private List<IColumn<Idea,String>> createColumns() {
        List<IColumn<Idea,String>> columns = new ArrayList<>();
        columns.add(new DateColumn<>(Model.of(HEADER_DATE_CREATED), "dateCreated", "dateCreated"));
        columns.add(new PropertyColumn<>(Model.of(HEADER_PROJECT_TYPE), "projectType.name", "projectType.name"));

        columns.add(new PropertyColumn<Idea, String>(Model.of(HEADER_TITLE), TITLE) {
            @Override
            public void populateItem(Item<ICellPopulator<Idea>> components, String s, final IModel<Idea> model) {
                components.add(new AjaxLinkPanel<Idea>(s, model, new PropertyModel(model, TITLE)) {
                    @Override
                    public void onClick(AjaxRequestTarget target, IModel<Idea> model) {
                        openDialog(model, target);
                    }
                });
            }
        });

        columns.add(new AbstractExportableColumn<Idea, String>(Model.of(HEADER_STATUS), "match.status") {
            @Override
            public IModel<String> getDataModel(IModel<Idea> rowModel) {
                return Model.of(ideaService.getStatusString(rowModel.getObject()));
            }
        });

        columns.add(new AbstractExportableColumn<Idea, String>(Model.of(HEADER_SUPERVISOR)) {
            @Override
            public void populateItem(Item<ICellPopulator<Idea>> item, String id, IModel<Idea> ideaModel) {
                addSupervisorField(item, id, ideaModel);
            }

            @Override
            public IModel<String> getDataModel(IModel<Idea> rowModel) {
                return new PropertyModel<>(rowModel, "match.supervisor.fullName");
            }
        });

        columns.add(new AbstractExportableColumn<Idea, String>(Model.of(HEADER_AUTHORS)) {
            @Override
            public void populateItem(Item<ICellPopulator<Idea>> cellItem, String componentId, IModel<Idea> rowModel) {
                cellItem.add(new DisplayMultiplesPanel<IdeaParticipation>(componentId, new ListAdapterModel<>(new PropertyModel<Iterable<IdeaParticipation>>(rowModel, "ideaParticipations"))) {
                    @Override
                    public Component getComponent(String componentId, IModel<IdeaParticipation> t) {
                        return new StudentProfileProgramPanel(componentId, t, true);
                    }
                });
            }

            @Override
            public IModel<String> getDataModel(IModel<Idea> rowModel) {
                StringBuilder participants = new StringBuilder();
                Set<IdeaParticipation> participantsSet = rowModel.getObject().getIdeaParticipations();
                for (IdeaParticipation participation: participantsSet) {
                    participants.append(participation.getUser().getFullName()).append("; ");
                }
                return Model.of(participants.toString());
            }
        });
        columns.addAll(addPanelSpecificColumns());
        return columns;
    }

    private void openDialog(IModel<Idea> model, AjaxRequestTarget target) {
        modalWindowPlus.setContent(new AdminIdeaDetailsPanel(modalWindowPlus.getContentId(), model, dataTable) {

            @Override
            public void update(AjaxRequestTarget target) {
                target.add(dataTable);
                modalWindowPlus.close(target);
                target.appendJavaScript("window.scrollTo(0,0);");
            }

        });
        modalWindowPlus.setTitle(model.getObject().getTitle());
        modalWindowPlus.show(target);
    }

    private void addSupervisorField(Item<ICellPopulator<Idea>> item, String id, final IModel<Idea> ideaModel) {
        switch (ideaModel.getObject().getType()) {
            case SUPERVISOR:
                item.add(new UserLinkPanel(id, new PropertyModel<>(ideaModel, "match.supervisor")));
                break;
            case STUDENT:
                item.add(new EmployeeAutoCompleteDivPanel(id, new PropertyModel<>(ideaModel, "match.supervisor")) {
                    @Override
                    public void onNewEmployeeSelection(AjaxRequestTarget target, User newSelection) {
                        Idea idea = ideaModel.getObject();
                        if (newSelection == null) {
                            ideaService.adminUnmatchIdea(idea, SciProSession.get().getUser());
                            info("Unmatched idea: " + idea.getTitle());
                        }
                        else {
                            if (targetService.hasTargetsLeft(idea.getApplicationPeriod(), newSelection, idea.getProjectType())) {
                                ideaService.changeSupervisor(idea, newSelection, SciProSession.get().getUser());
                                info("Supervisor changed");
                            } else {
                                error("The supervisor have reached current target numbers");
                            }
                        }
                        target.addListener(new AjaxFeedbackPanelUpdater());
                    }

                    @Override
                    protected void onConfigure() {
                        super.onConfigure();
                        setEnabled(ideaModel.getObject().getProject() == null);
                    }

                    @Override
                    protected boolean allowClear() {
                        return true;
                    }
                });
                break;
            default:
                throw new IllegalStateException("idea needs a type");
        }
    }

    private void addFiltering() {
        final Long applicationPeriodId = getSession().getMetaData(PRESELECTED_APPLICATION_PERIOD);
        if (applicationPeriodId != null) {
            filter.setApplicationPeriod(applicationPeriodService.findOne(applicationPeriodId));
        }
        filter.setTypes(new ArrayList<>(asList(Idea.Type.STUDENT, Idea.Type.SUPERVISOR)));
        filter.setStatuses(new ArrayList<>(asList(Idea.Status.UNMATCHED, Idea.Status.MATCHED)));
        filter.setLevels(projectTypeService.findWithModule(ProjectModule.MATCH));
        add(new AjaxIdeaFilterPanel("filter", Model.of(filter)) {
            @Override
            protected void filterChanged(final AjaxRequestTarget target) {
                target.add(dataTable);
            }
        });
    }

    protected abstract List<IColumn<Idea, String>> addPanelSpecificColumns();
}
