package se.su.dsv.scipro.match;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.ajax.form.AjaxFormComponentUpdatingBehavior;
import org.apache.wicket.ajax.markup.html.form.AjaxCheckBox;
import org.apache.wicket.markup.html.form.ChoiceRenderer;
import org.apache.wicket.markup.html.form.EnumChoiceRenderer;
import org.apache.wicket.markup.html.form.TextField;
import org.apache.wicket.markup.html.panel.GenericPanel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.LoadableDetachableModel;
import org.apache.wicket.model.Model;
import org.apache.wicket.model.PropertyModel;
import org.joda.time.DateTime;
import se.su.dsv.scipro.components.*;
import se.su.dsv.scipro.date.DeactivatableDatePickerPanel;
import se.su.dsv.scipro.springdata.services.UnitService;
import se.su.dsv.scipro.system.*;

import javax.inject.Inject;
import java.util.*;

import static java.util.Arrays.asList;
import static se.su.dsv.scipro.match.AbstractAdminIdeaPanel.PRESELECTED_APPLICATION_PERIOD;

public abstract class AjaxIdeaFilterPanel extends GenericPanel<IdeaService.Filter> {
    private static final int MIN_TITLE_LENGTH = 3;

    @Inject
    private ProjectTypeService projectTypeService;
    @Inject
    private UnitService unitService;
    @Inject
    private KeywordService keywordService;
    @Inject
    private ResearchAreaService researchAreaService;
    @Inject
    private ApplicationPeriodService applicationPeriodService;

    private KeywordFilterPanel keywordPanel;

    public AjaxIdeaFilterPanel(final String id, final IModel<IdeaService.Filter> model) {
        super(id, model);
        statusFilter();
        ideaTypeFilter();
        projectTypeFilter();
        unitFilter();
        newKeywordsFilter();
        researchAreaFilter();
        newSupervisorFilter();
        newAuthorFilter();
        programFilter();
        dateFilter();
        add(newTitleFilter("titleFilter"));
        newApplicationPeriodFilter();
        addFirstMeetingFilter();
        addPublishedFilter();
    }
    
    protected abstract void filterChanged(AjaxRequestTarget target);

    private void statusFilter() {
        add(new AjaxCheckBoxMultipleChoice<Idea.Status>(
                "statusFilter",
                new PropertyModel<List<Idea.Status>>(getModel(), "statuses"),
                getStatusChoices(),
                new EnumChoiceRenderer<>(AjaxIdeaFilterPanel.this)) {
            @Override
            public void onUpdate(AjaxRequestTarget target) {
                filterChanged(target);
            }
        });
    }

    private List<Idea.Status> getStatusChoices() {
        return asList(Idea.Status.values());
    }

    private void ideaTypeFilter() {
        add(new AjaxCheckBoxMultipleChoice<Idea.Type>(
                "typeFilter",
                new PropertyModel<List<Idea.Type>>(getModel(), "types"),
                asList(Idea.Type.STUDENT, Idea.Type.SUPERVISOR),
                new EnumChoiceRenderer<>(AjaxIdeaFilterPanel.this)) {
            @Override
            public void onUpdate(AjaxRequestTarget target) {
                filterChanged(target);
            }
        });
    }

    private void projectTypeFilter() {
        add(new AjaxCheckBoxMultipleChoice<ProjectType>(
                "projectTypeFilter",
                new PropertyModel<List<ProjectType>>(getModel(), "levels"),
                projectTypeService.findWithModule(ProjectModule.MATCH),
                new ChoiceRenderer<>("name", "id")) {
            @Override
            public void onUpdate(AjaxRequestTarget target) {
                filterChanged(target);
            }
        });
    }

    private void unitFilter() {
        add(new AjaxDropDownChoice<Unit>(
                "unitFilter",
                new PropertyModel<>(getModel(), "unit"),
                unitService.findAll(),
                new ChoiceRenderer<>("title", "id")) {
            @Override
            public void onNewSelection(AjaxRequestTarget target, Unit objectSelected) {
                filterChanged(target);
            }
        });
    }

    private void newKeywordsFilter() {
        keywordPanel = new KeywordFilterPanel("keywordsFilter", getKeywords()) {
            @Override
            public void onKeywordsSelected(AjaxRequestTarget target, Set<Keyword> keywords) {
                AjaxIdeaFilterPanel.this.getModelObject().setKeywords(keywords);
                filterChanged(target);
            }
        };
        keywordPanel.setVisible(false);
        keywordPanel.setOutputMarkupPlaceholderTag(true);
        add(keywordPanel);
    }

    private IModel<Set<Keyword>> getKeywords() {
        return new LoadableDetachableModel<Set<Keyword>>() {
            @Override
            protected Set<Keyword> load() {
                final ResearchArea researchArea = getModelObject().getResearchArea();
                if (researchArea != null) {
                    return new HashSet<>(keywordService.findByResearchArea(researchArea));
                }
                return new HashSet<>();
            }
        };
    }

    private void researchAreaFilter() {
        add(new AjaxDropDownChoice<ResearchArea>(
                "areaFilter",
                new PropertyModel<>(getModel(), "researchArea"),
                researchAreaService.findNotDeleted(),
                new ChoiceRenderer<>("title", "id")) {
            @Override
            public void onNewSelection(AjaxRequestTarget target, ResearchArea objectSelected) {
                filterChanged(target);
                keywordPanel.setVisible(objectSelected != null && !keywordService.findByResearchArea(objectSelected).isEmpty());
                target.add(keywordPanel);
            }
        });
    }

    private void newSupervisorFilter() {
        add(new EmployeeAutoComplete("supervisorFilter", new PropertyModel<>(getModel(), "supervisor")) {
            @Override
            protected void action(AjaxRequestTarget target, User employee) {
                filterChanged(target);
            }
        });
    }

    private void newAuthorFilter() {
        add(new AuthorAutoComplete("authorFilter", new PropertyModel<>(getModel(), "author")) {
            @Override
            protected void action(AjaxRequestTarget target, User newSelection) {
                filterChanged(target);
            }
        });
    }

    private void programFilter() {
        add(new ProgramAutoComplete("programFilter", new PropertyModel<>(getModel(), "program")) {
            @Override
            protected void action(AjaxRequestTarget target, Program object) {
                filterChanged(target);
            }
        });
    }

    private void dateFilter() {
        Date startDate = new DateTime().minusYears(1).toDate();
        add(new DeactivatableDatePickerPanel("dateContainer", Model.of(startDate), new Model<>(), Model.of(false)) {
            @Override
            protected void datesChanged(final AjaxRequestTarget target) {
                getModelObject().setCreatedAfter(getStartDate().getObject());
                getModelObject().setCreatedBefore(getEndDate().getObject());
                filterChanged(target);
            }
        });
    }

    private TextField<String> newTitleFilter(final String id) {
        TextField<String> titleField = new TextField<>(id, new Model<>());
        titleField.add(new AjaxFormComponentUpdatingBehavior("keyup") {
            @Override
            protected void onUpdate(AjaxRequestTarget target) {
                final String titleString = titleField.getModelObject();
                if (titleString != null && titleString.length() >= MIN_TITLE_LENGTH) {
                    getModelObject().setTitleContains(titleString);
                } else {
                    getModelObject().setTitleContains(null);
                }
                filterChanged(target);
            }
        });
        return titleField;
    }

    private void newApplicationPeriodFilter() {
        List<ApplicationPeriod> list = applicationPeriodService.findAll();
        Collections.sort(list, Comparator.comparing(ApplicationPeriod::getStartDate).reversed());
        final AjaxDropDownChoice<ApplicationPeriod> applicationPeriodAjaxDropDownChoice = new AjaxDropDownChoice<ApplicationPeriod>(
                "applicationPeriodFilter",
                new PropertyModel<>(getModel(), "applicationPeriod"),
                list,
                new ChoiceRenderer<>("name", "id")) {
            @Override
            public void onNewSelection(AjaxRequestTarget target, ApplicationPeriod objectSelected) {
                filterChanged(target);
                getSession().setMetaData(PRESELECTED_APPLICATION_PERIOD, objectSelected != null ? objectSelected.getId() : null);
            }
        };
        applicationPeriodAjaxDropDownChoice.setNullValid(true);
        add(applicationPeriodAjaxDropDownChoice);
    }

    private void addFirstMeetingFilter() {
        add(new AjaxCheckBox("noFirstMeeting", new PropertyModel<>(getModel(), "noFirstMeeting")) {
            @Override
            protected void onUpdate(final AjaxRequestTarget target) {
                filterChanged(target);
            }
        });
    }

    private void addPublishedFilter() {
        add(new AjaxRadioChoice<PublishedFilter>("publishedFilter", new Model<>(PublishedFilter.ALL), Arrays.asList(PublishedFilter.values()), new EnumChoiceRenderer<>(this)) {
            {
                setRequired(true);
            }

            @Override
            protected void update(final AjaxRequestTarget target) {
                switch (getModelObject()) {
                    case ALL:
                        AjaxIdeaFilterPanel.this.getModelObject().setPublished(null);
                        break;
                    case PUBLISHED:
                        AjaxIdeaFilterPanel.this.getModelObject().setPublished(Boolean.TRUE);
                        break;
                    case UNPUBLISHED:
                        AjaxIdeaFilterPanel.this.getModelObject().setPublished(Boolean.FALSE);
                        break;
                }
                filterChanged(target);
            }
        });
    }

    private enum PublishedFilter { ALL, PUBLISHED, UNPUBLISHED }
}
