package se.su.dsv.scipro.match;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.extensions.markup.html.form.palette.Palette;
import org.apache.wicket.extensions.markup.html.form.palette.theme.DefaultTheme;
import org.apache.wicket.markup.html.WebMarkupContainer;
import org.apache.wicket.markup.html.form.*;
import org.apache.wicket.markup.html.panel.FeedbackPanel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.LoadableDetachableModel;
import org.apache.wicket.model.Model;
import org.apache.wicket.util.collections.ArrayListStack;
import se.su.dsv.scipro.components.AjaxDropDownChoice;
import se.su.dsv.scipro.components.CoAuthorChoice;
import se.su.dsv.scipro.components.SupervisorAutoComplete;
import se.su.dsv.scipro.data.DetachableServiceModel;
import se.su.dsv.scipro.data.DetachableServiceModelCollection;
import se.su.dsv.scipro.system.*;

import javax.inject.Inject;
import java.util.*;

import static java.util.Comparator.comparing;
import static java.util.Comparator.reverseOrder;
import static se.su.dsv.scipro.match.AbstractAdminIdeaPanel.PRESELECTED_APPLICATION_PERIOD;

public class ManualMatching extends AbstractAdminMatchPage {
    @Inject
    private ResearchAreaService researchAreaService;
    @Inject
    private KeywordService keywordService;
    @Inject
    private UserService userService;
    @Inject
    private ProjectTypeService projectTypeService;
    @Inject
    private ApplicationPeriodService applicationPeriodService;
    @Inject
    private IdeaService ideaService;

    public ManualMatching() {
        add(new ManualMatchingForm("form"));
    }

    private class ManualMatchingForm extends Form<Idea> {
        private final IModel<ApplicationPeriod> applicationPeriod =
                new DetachableServiceModel<>(applicationPeriodService, getSession().getMetaData(PRESELECTED_APPLICATION_PERIOD));
        private final IModel<ResearchArea> researchArea = new DetachableServiceModel<>(researchAreaService);
        private final IModel<Collection<Keyword>> keywords = new DetachableServiceModelCollection<>(keywordService);
        private final IModel<User> supervisor = new DetachableServiceModel<>(userService);
        private final IModel<Collection<User>> authors = new DetachableServiceModelCollection<>(userService);
        private final IModel<String> title = new Model<>();
        private final IModel<String> description = new Model<>();
        private final IModel<ProjectType> projectType = new DetachableServiceModel<>(projectTypeService);
        private final IModel<Set<Language>> languages = Model.ofSet(EnumSet.noneOf(Language.class));

        private ManualMatchingForm(final String id) {
            super(id);

            add(new FeedbackPanel("feedback"));

            final AjaxDropDownChoice<ApplicationPeriod> applicationPeriodChoice =
                    new AjaxDropDownChoice<ApplicationPeriod>(
                            "applicationPeriod",
                            applicationPeriod,
                            getApplicationPeriods(),
                            new ChoiceRenderer<>("name", "id"))
                    {
                        @Override
                        public void onNewSelection(final AjaxRequestTarget target, final ApplicationPeriod ap) {
                            getSession().setMetaData(PRESELECTED_APPLICATION_PERIOD, ap != null ? ap.getId() : null);
                        }
                    };
            applicationPeriodChoice.setRequired(true);
            add(applicationPeriodChoice);

            final CheckBoxMultipleChoice<Language> languageChoice =
                    new CheckBoxMultipleChoice<>("languages", this.languages, Arrays.asList(Language.values()), new EnumChoiceRenderer<>(this));
            languageChoice.setLabelPosition(AbstractChoice.LabelPosition.WRAP_AFTER);
            languageChoice.setPrefix("<div class=\"checkbox\">");
            languageChoice.setSuffix("</div>");
            languageChoice.setRequired(true);
            add(languageChoice);

            final RadioChoice<ProjectType> projectTypeChoice =
                    new RadioChoice<>("projectType", this.projectType, getProjectTypes(), new ChoiceRenderer<>("name", "id"));
            projectTypeChoice.setLabelPosition(AbstractChoice.LabelPosition.WRAP_AFTER);
            projectTypeChoice.setPrefix("<div class=\"radio\">");
            projectTypeChoice.setSuffix("</div>");
            projectTypeChoice.setRequired(true);
            add(projectTypeChoice);

            add(new RequiredTextField<>("title", title, String.class));

            add(new TextArea<>("description", description));

            final SupervisorAutoComplete supervisorChoice = new SupervisorAutoComplete("supervisor", this.supervisor);
            supervisorChoice.setRequired(true);
            add(supervisorChoice);

            add(new CoAuthorChoice("authors", authors));

            final WebMarkupContainer keywordContainer = new WebMarkupContainer("keywordContainer");
            keywordContainer.setOutputMarkupId(true);

            final AjaxDropDownChoice<ResearchArea> researchAreaChoice =
                    new AjaxDropDownChoice<ResearchArea>(
                            "researchAreas",
                            researchArea,
                            getResearchAreas(),
                            new ChoiceRenderer<>("title", "id"))
                    {
                        @Override
                        public void onNewSelection(final AjaxRequestTarget target, final ResearchArea objectSelected) {
                            target.add(keywordContainer);
                        }
                    };
            researchAreaChoice.setRequired(true);
            add(researchAreaChoice);

            final Palette<Keyword> keywordPalette = new Palette<Keyword>("keywords", keywords, getKeywords(researchArea), new ChoiceRenderer<>("keyword", "id"), 10, false) {
                @Override
                protected void onConfigure() {
                    super.onConfigure();
                    setVisibilityAllowed(researchArea.getObject() != null);
                }
            };
            keywordPalette.setRequired(true);
            keywordPalette.add(new DefaultTheme());
            keywordContainer.add(keywordPalette);
            add(keywordContainer);
        }

        @Override
        protected void onSubmit() {
            final Idea idea = new Idea();
            idea.setType(Idea.Type.SUPERVISOR);
            idea.setProjectType(projectType.getObject());
            idea.setTitle(title.getObject());
            idea.setApplicationPeriod(applicationPeriod.getObject());
            idea.setCreator(supervisor.getObject());
            idea.setKeywords(new HashSet<>(keywords.getObject()));
            idea.setLanguages(languages.getObject());
            idea.setPublished(true);
            idea.setResearchArea(researchArea.getObject());
            idea.setDescription(description.getObject());
            for (User author : authors.getObject()) {
                idea.addIdeaParticipation(new IdeaParticipation(author, idea));
            }
            ideaService.saveSupervisorIdea(
                    idea,
                    supervisor.getObject(),
                    new ArrayListStack<>(keywords.getObject()),
                    true);
            getSession().success("Idea created");
            setResponsePage(ManualMatching.class);
        }
    }

    private IModel<List<ApplicationPeriod>> getApplicationPeriods() {
        return new LoadableDetachableModel<List<ApplicationPeriod>>() {
            @Override
            protected List<ApplicationPeriod> load() {
                final List<ApplicationPeriod> all = applicationPeriodService.findAll();
                Collections.sort(all, comparing(ApplicationPeriod::getStartDate, reverseOrder()));
                return all;
            }
        };
    }

    private IModel<List<ProjectType>> getProjectTypes() {
        return new LoadableDetachableModel<List<ProjectType>>() {
            @Override
            protected List<ProjectType> load() {
                return projectTypeService.findWithModule(ProjectModule.MATCH);
            }
        };
    }

    private IModel<List<Keyword>> getKeywords(final IModel<ResearchArea> researchArea) {
        return new LoadableDetachableModel<List<Keyword>>() {
            @Override
            protected List<Keyword> load() {
                return keywordService.findByResearchArea(researchArea.getObject());
            }
        };
    }

    private IModel<List<ResearchArea>> getResearchAreas() {
        return new LoadableDetachableModel<List<ResearchArea>>() {
            @Override
            protected List<ResearchArea> load() {
                return researchAreaService.findNotDeleted();
            }
        };
    }
}
