package se.su.dsv.scipro.match;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.markup.html.WebMarkupContainer;
import org.apache.wicket.markup.html.form.*;
import org.apache.wicket.markup.html.link.BookmarkablePageLink;
import org.apache.wicket.markup.html.panel.ComponentFeedbackPanel;
import org.apache.wicket.markup.html.panel.FeedbackPanel;
import org.apache.wicket.markup.html.panel.GenericPanel;
import org.apache.wicket.model.*;
import org.apache.wicket.model.util.ListModel;
import org.apache.wicket.model.util.SetModel;
import org.apache.wicket.request.mapper.parameter.PageParameters;
import org.apache.wicket.validation.validator.StringValidator;
import se.su.dsv.scipro.components.*;
import se.su.dsv.scipro.project.ProjectService;
import se.su.dsv.scipro.projectpartner.ProjectPartnerPage;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.*;
import se.su.dsv.scipro.util.PageParameterKeys;

import javax.inject.Inject;
import java.util.*;
import java.util.stream.Collectors;

public class ProjectIdeaSubmissionPanel extends GenericPanel<Idea> {

    public static final String PROJECT_TYPE = "projectType";
    public static final String CO_AUTHOR_CHOICE = "coAuthorChoice";
    public static final String IDEA_FORM = "ideaForm";
    public static final String PROGRAM = "program";
    public static final String PROGRAM_DROP_DOWN = "programDropDown";
    @Inject
    private IdeaService ideaService;
    @Inject
    private ResearchAreaService researchAreaService;
    @Inject
    private KeywordService keywordService;
    @Inject
    private ApplicationPeriodService applicationPeriodService;
    @Inject
    private UserService userService;
    @Inject
    private ProjectService projectService;

    private PageParameters pp;

    private Set<Keyword> keywords = new HashSet<>();

    public ProjectIdeaSubmissionPanel(String id, IModel<Idea> model, boolean isNewIdea) {
        super(id, model);

        pp = new PageParameters();
        pp.add(PageParameterKeys.MAP.get(ApplicationPeriod.class), model.getObject().getApplicationPeriod().getId());

        IdeaForm form = new IdeaForm(IDEA_FORM, model, isNewIdea);
        add(form);
        add(new BookmarkablePageLink<Void>("sitop", ProjectNewAllSupervisorIdeasPage.class, pp));
        addFeedback(form);
        addDialog();
    }

    private void addDialog() {
        ModalWindowPlus modalWindowPlus = new ModalWindowPlus("dialog");
        add(modalWindowPlus);
    }

    private void addFeedback(IdeaForm form) {
        FeedbackPanel feedback = new ComponentFeedbackPanel("feedback", form);
        feedback.setOutputMarkupId(true);
        add(feedback);
    }

    private LoadableDetachableModel<User> getLoadedStudent(){
        return new LoadableDetachableModel<User>() {
            @Override
            protected User load() {
                return userService.findOne(SciProSession.get().getUser().getId());
            }
        };
    }

    private class IdeaForm extends Form<Idea> {

        public static final int MIN_TITLE_LENGTH = 5;
        private static final int MAX_KEYWORDS_SIZE = 5;

        private final boolean isNewIdea;
        private KeywordFilterPanel keywordsPanel;
        private DropDownChoice<Program> programDropDownChoice;
        private User creator;
        private CoAuthorChoice coAuthorChoice;
        private RadioChoice<ProjectType> projectTypeChoice;

        public IdeaForm(String id, IModel<Idea> model, final boolean isNewIdea) {
            super(id, model);
            this.isNewIdea = isNewIdea;
            this.creator = SciProSession.get().getUser();
            addProjectTypeSelection();
            addLanguageSelection();
            addCoAuthorSelection();
            addProgramSelection();
            addTitleField();
            addTholanderFields();
            addResearchAreaAndKeywordsSelection();
            addContract();
        }

        private void addProjectTypeSelection() {
            projectTypeChoice = new RadioChoice<>(PROJECT_TYPE, new PropertyModel<>(getModel(), "projectType"),
                    new ArrayList<>(getModelObject().getApplicationPeriod().getProjectTypes()), new ChoiceRenderer<>("name", "id"));
            projectTypeChoice.setRequired(true);
            projectTypeChoice.setLabelPosition(AbstractChoice.LabelPosition.WRAP_AFTER);
            projectTypeChoice.setPrefix("<div class=\"radio\">");
            projectTypeChoice.setSuffix("</div>");
            add(projectTypeChoice);
            add(new ComponentFeedbackPanel("projectTypeFeedback", projectTypeChoice));
        }

        private void addCoAuthorSelection() {
            final Set<User> partners = getModelObject()
                    .getIdeaParticipations()
                    .stream()
                    .map(IdeaParticipation::getUser)
                    .filter(user -> !user.equals(SciProSession.get().getUser()))
                    .collect(Collectors.toSet());
            coAuthorChoice = new CoAuthorChoice(CO_AUTHOR_CHOICE, Model.of(partners));
            add(coAuthorChoice);
            add(new ComponentFeedbackPanel("coAuthorFeedback", coAuthorChoice));
            add(new BookmarkablePageLink<Void>("partnerLink", ProjectPartnerPage.class));
        }

        private void addProgramSelection() {
            final User student = getLoadedStudent().getObject();
            WebMarkupContainer program = new WebMarkupContainer(PROGRAM) {
                @Override
                protected void onConfigure() {
                    setVisibilityAllowed(student.hasPrograms());
                }
            };
            add(program);
            programDropDownChoice = new DropDownChoice<>(PROGRAM_DROP_DOWN, new Model<>(),
                    new ArrayList<>(student.getPrograms()), new ChoiceRenderer<>("name", "id"));
            programDropDownChoice.setNullValid(true);
            program.add(programDropDownChoice);
        }


        private void addLanguageSelection() {
            CheckBoxMultipleChoice<Language> language = new CheckBoxMultipleChoice<>("languages", new PropertyModel<Set<Language>>(getModel(), "languages"),
                    Arrays.asList(Language.values()), new EnumChoiceRenderer<>(this));
            language.setPrefix("<div class=\"checkbox\">");
            language.setSuffix("</div>");
            language.setRequired(true);
            language.setLabelPosition(AbstractChoice.LabelPosition.WRAP_AFTER);
            add(new ComponentFeedbackPanel("languageFeedback", language));
            add(language);
            add(new InfoPanel("languageInfo", new ResourceModel("languageInfo")));
        }

        private void addTitleField() {
            TextField<String> title = new TextField<>("title", new PropertyModel<String>(getModel(), "title"));
            title.setRequired(true);
            title.add(StringValidator.minimumLength(MIN_TITLE_LENGTH));
            add(new ComponentFeedbackPanel("titleFeedback", title));
            add(title);
        }

        private void addTholanderFields() {
            CounterTextArea what = new CounterTextArea("background", new PropertyModel<>(getModel(), "tholanderBox.background"), TholanderBox.MAX_CHARS);
            what.setRequired(true);
            add(what);

            CounterTextArea why = new CounterTextArea("literature", new PropertyModel<>(getModel(), "tholanderBox.literature"), TholanderBox.MAX_CHARS);
            why.setRequired(true);
            add(why);

            CounterTextArea theory = new CounterTextArea("problem", new PropertyModel<>(getModel(), "tholanderBox.problem"), TholanderBox.MAX_CHARS);
            theory.setRequired(true);
            add(theory);

            CounterTextArea practical = new CounterTextArea("method", new PropertyModel<>(getModel(), "tholanderBox.method"), TholanderBox.MAX_CHARS);
            practical.setRequired(true);
            add(practical);

            final CounterTextArea interests = new CounterTextArea("interests", new PropertyModel<>(getModel(), "tholanderBox.interests"), TholanderBox.MAX_CHARS);
            interests.setRequired(true);
            add(interests);
        }

        private void addResearchAreaAndKeywordsSelection() {
            final WebMarkupContainer areaContainer = new WebMarkupContainer("areaContainer");
            areaContainer.setOutputMarkupId(true);

            List<ResearchArea> areas = researchAreaService.findNotDeleted();

            AjaxDropDownChoice<ResearchArea> areaDropDown = new AjaxDropDownChoice<ResearchArea>("area", new PropertyModel<>(getModel(), "researchArea"),
                    new ListModel<>(areas), new ChoiceRenderer<>("title", "id")) {
                @Override
                public void onNewSelection(AjaxRequestTarget target, ResearchArea selected) {
                    Set<Keyword> keywordHashSet = new HashSet<>(keywordService.findByResearchArea(selected));
                    keywordsPanel.setModelObject(keywordHashSet);
                    keywordsPanel.setVisible(!keywordHashSet.isEmpty());
                    target.add(areaContainer);
                    target.add(keywordsPanel);
                }
            };
            areaDropDown.setRequired(true);
            areaContainer.add(areaDropDown);
            areaContainer.add(new ComponentFeedbackPanel("researchAreaFeedback", areaDropDown));

            final Set<Keyword> keywords = getModelObject().getResearchArea() == null
                    ? new HashSet<>()
                    : new HashSet<>(keywordService.findByResearchArea(getModelObject().getResearchArea()));
            ProjectIdeaSubmissionPanel.this.keywords = getModelObject().getKeywords();
            keywordsPanel = new KeywordFilterPanel("keywords", new SetModel<>(keywords), new ListModel<>(new ArrayList<>(getModelObject().getKeywords()))) {
                @Override
                public void onKeywordsSelected(AjaxRequestTarget target, Set<Keyword> selected) {
                    ProjectIdeaSubmissionPanel.this.keywords = selected;
                }
            };
            keywordsPanel.setOutputMarkupPlaceholderTag(true);
            keywordsPanel.setVisible(getModelObject().getResearchArea() != null);
            areaContainer.add(keywordsPanel);
            areaContainer.add(new ComponentFeedbackPanel("keywordsFeedback", keywordsPanel));
            add(areaContainer);
        }

        private void addContract() {
            CheckBox contract = new CheckBox("contract", new Model<>());
            contract.setModelObject(!isNewIdea);
            contract.add(new CheckBoxRequiredValidator());
            add(contract);
            add(new ComponentFeedbackPanel("contractFeedback", contract));
            add(new BookmarkablePageLink<Void>("sibottom", ProjectNewAllSupervisorIdeasPage.class, pp));
        }

        @Override
        protected void onSubmit() {
            ideaService.saveStudentIdea(getModelObject(), creator, programDropDownChoice.getModelObject(), new HashSet<>(coAuthorChoice.getModelObject()),
                    new ArrayList<>(keywords), isNewIdea);
            setResponsePage(ProjectIdeaStartPage.class);
        }


        @Override
        protected void onError() {
            if (getFeedbackMessages().isEmpty()) {
                error(getString("submissionFailed"));
            }
        }

        @Override
        protected void onValidate() {
            super.onValidate();
            if (keywordNotWithinLimits()) {
                keywordsPanel.error(getString("keywordError"));
            }
            final Collection<User> partners = new HashSet<>(coAuthorChoice.getConvertedInput());
            for (User partner : partners) {
                if (ideaService.authorParticipatingOnOtherActiveIdea(partner, getModelObject().getApplicationPeriod(), getModelObject())) {
                    coAuthorChoice.error(getString("coAuthorParticipating", Model.of(partner)));
                }
            }
            if (ideaService.authorParticipatingOnOtherActiveIdea(SciProSession.get().getUser(), getModelObject().getApplicationPeriod(), getModelObject())) {
                error(getString("you.are.partner.on.other.idea"));
            }
            if (partners.contains(SciProSession.get().getUser())) {
                coAuthorChoice.error(getString("addedYourself"));
            }
            final ProjectType projectType = projectTypeChoice.getConvertedInput();
            if (projectType != null) {
                if (!projectService.getActiveProjects(SciProSession.get().getUser(), projectType).isEmpty()) {
                    error(getString("you.already.have.an.active.project.on.this.level"));
                }
                for (User partner : partners) {
                    if (!projectService.getActiveProjects(partner, projectType).isEmpty()) {
                        error(getString("partner.already.has.an.active.project.on.this.level", Model.of(partner)));
                    }
                }
                if (!applicationPeriodService.hasExemption(SciProSession.get().getUser(), getModelObject().getApplicationPeriod())) {
                    if (partners.size() < projectType.getMinAuthors() - 1) {
                        coAuthorChoice.error(getString("too.few.authors", Model.of(projectType)));
                    }
                    if (partners.size() > projectType.getMaxAuthors() - 1) {
                        coAuthorChoice.error(getString("too.many.authors", Model.of(projectType)));
                    }
                }
            }
        }

        private boolean keywordNotWithinLimits() {
            return keywords.isEmpty() || keywords.size() > MAX_KEYWORDS_SIZE;
        }
    }
}
