package se.su.dsv.scipro.repository.panels;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.extensions.markup.html.repeater.data.grid.ICellPopulator;
import org.apache.wicket.extensions.markup.html.repeater.data.sort.SortOrder;
import org.apache.wicket.extensions.markup.html.repeater.data.table.AbstractColumn;
import org.apache.wicket.extensions.markup.html.repeater.data.table.IColumn;
import org.apache.wicket.extensions.markup.html.repeater.data.table.PropertyColumn;
import org.apache.wicket.extensions.markup.html.repeater.util.SortableDataProvider;
import org.apache.wicket.markup.html.form.Form;
import org.apache.wicket.markup.html.form.upload.FileUpload;
import org.apache.wicket.markup.html.form.upload.FileUploadField;
import org.apache.wicket.markup.html.panel.FeedbackPanel;
import org.apache.wicket.markup.html.panel.Fragment;
import org.apache.wicket.markup.html.panel.GenericPanel;
import org.apache.wicket.markup.repeater.Item;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.Model;
import org.apache.wicket.model.PropertyModel;
import org.apache.wicket.util.lang.Bytes;
import se.su.dsv.scipro.components.AjaxIconLinkPanel;
import se.su.dsv.scipro.components.DateColumn;
import se.su.dsv.scipro.components.EnumPropertyColumn;
import se.su.dsv.scipro.components.ExportableDataPanel;
import se.su.dsv.scipro.data.DetachableServiceModel;
import se.su.dsv.scipro.data.enums.DateStyle;
import se.su.dsv.scipro.dataproviders.PageAdapter;
import se.su.dsv.scipro.file.*;
import se.su.dsv.scipro.files.WicketProjectFileUpload;
import se.su.dsv.scipro.notifications.NotificationController;
import se.su.dsv.scipro.notifications.dataobject.NotificationSource;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.repository.components.FileDownloadLink;
import se.su.dsv.scipro.security.auth.Authorization;
import se.su.dsv.scipro.session.SciProSession;

import javax.inject.Inject;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import static se.su.dsv.scipro.notifications.dataobject.ProjectEvent.Event.FILE_UPLOADED;
import static se.su.dsv.scipro.security.auth.roles.Roles.AUTHOR;
import static se.su.dsv.scipro.security.auth.roles.Roles.SUPERVISOR;

@Authorization(authorizedRoles = {AUTHOR, SUPERVISOR})
public class ProjectFilePanel extends GenericPanel<Project> {

    public static final String FEEDBACK_PANEL = "feedbackPanel";
    public static final String AJAX_SIMPLE_UPLOAD = "ajax-simpleUpload";
    public static final String FILE_INPUT = "fileInput";
    private static final Bytes MAX_FILE_SIZE = Bytes.megabytes(100);
    public static final String TABLE = "table";
    @Inject
    private NotificationController notificationController;
    @Inject
    private ProjectFileService projectFileService;
    private FeedbackPanel feedback;
    private ExportableDataPanel<ProjectFile, String> dp;

    public ProjectFilePanel(String id, IModel<Project> project) {
        super(id, project);
    }

    @Override
    protected void onInitialize() {
        super.onInitialize();

        feedback = new FeedbackPanel(FEEDBACK_PANEL);
        feedback.setOutputMarkupId(true);
        add(feedback);

        final FileUploadForm fileUploadForm = new FileUploadForm(AJAX_SIMPLE_UPLOAD);
        add(fileUploadForm);

        dp = new ExportableDataPanel<>(TABLE, getColumns(), new FileProvider());
        dp.setOutputMarkupId(true);
        add(dp);
    }

    private List<IColumn<ProjectFile, String>> getColumns() {
        List<IColumn<ProjectFile, String>> columns = new ArrayList<>();
        columns.add(nameColumn());
        columns.add(new DateColumn<>(Model.of("Created"), "dateCreated", "dateCreated", DateStyle.DATETIME));
        columns.add(new PropertyColumn<>(Model.of("Size"), "fileDescription.size", "fileDescription.formattedSize"));
        columns.add(new PropertyColumn<>(Model.of("Uploader"), "fileDescription.uploader.firstName", "fileDescription.uploader.fullName"));
        columns.add(new EnumPropertyColumn<>(Model.of("Source"), "fileSource", "fileSource"));
        columns.add(new AbstractColumn<ProjectFile, String>(Model.of("Delete")) {
            @Override
            public void populateItem(Item<ICellPopulator<ProjectFile>> cellItem, String componentId, final IModel<ProjectFile> rowModel) {

                cellItem.add(new AjaxIconLinkPanel<ProjectFile>(componentId, rowModel, Model.of(AjaxIconLinkPanel.DELETE), "Are you sure you want to delete the file?") {
                    @Override
                    public void onClick(AjaxRequestTarget target, IModel<ProjectFile> model) {
                        projectFileService.delete(rowModel.getObject());
                        success(getString("deleted"));
                        target.add(feedback);
                        target.add(dp);
                    }

                    @Override
                    protected void onConfigure() {
                        ProjectFile file = rowModel.getObject();
                        setVisibilityAllowed(file.getFileDescription().getUploader().equals(SciProSession.get().getUser()));
                    }
                });
            }
        });
        return columns;
    }

    private IColumn<ProjectFile, String> nameColumn() {
        return new AbstractColumn<ProjectFile, String>(Model.of("Name"), "fileDescription.name") {
            @Override
            public void populateItem(final Item<ICellPopulator<ProjectFile>> cellItem, final String componentId, final IModel<ProjectFile> rowModel) {
                FileDownloadLink<FileDescription> downloadLink = new FileDownloadLink<>("name", new PropertyModel<>(rowModel, "fileDescription"));
                downloadLink.setBody(new PropertyModel<>(rowModel, "fileDescription.name"));

                Fragment cell = new Fragment(componentId, "_name", ProjectFilePanel.this);
                cell.add(downloadLink);
                cellItem.add(cell);
            }
        };
    }

    public void notifyUpload(String filename) {
        Project project = getModelObject();
        NotificationSource source = new NotificationSource();
        source.setMessage(filename);
        notificationController.notifyProject(project, FILE_UPLOADED, source);
    }

    private class FileUploadForm extends Form<Void> {
        private FileUploadField fileUploadField;

        public FileUploadForm(String id) {
            super(id);
            setMultiPart(true);

            fileUploadField = new FileUploadField(FILE_INPUT);
            fileUploadField.setRequired(true);
            add(fileUploadField);

            setMaxSize(MAX_FILE_SIZE);
        }

        @Override
        protected void onSubmit() {
            final FileUpload upload = fileUploadField.getFileUpload();
            projectFileService.store(new WicketProjectFileUpload(upload, ProjectFilePanel.this.getModelObject()), FileSource.FILES);
            success(getString("file.uploaded"));
            notifyUpload(upload.getClientFileName());
        }
    }

    private final class FileProvider extends SortableDataProvider<ProjectFile, String> {

        private FileProvider() {
            setSort("dateCreated", SortOrder.DESCENDING);
        }

        @Override
        public Iterator<? extends ProjectFile> iterator(long first, long count) {
            return projectFileService.getProjectFiles(getModelObject(), new PageAdapter(first, count, getSort())).iterator();
        }

        @Override
        public long size() {
            return projectFileService.countProjectFiles(getModelObject());
        }

        @Override
        public IModel<ProjectFile> model(ProjectFile object) {
            return new DetachableServiceModel<>(projectFileService, object);
        }

    }
}
