package se.su.dsv.scipro.reviewer;

import org.apache.wicket.Component;
import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.extensions.markup.html.repeater.data.grid.ICellPopulator;
import org.apache.wicket.extensions.markup.html.repeater.data.table.DefaultDataTable;
import org.apache.wicket.extensions.markup.html.repeater.data.table.IColumn;
import org.apache.wicket.extensions.markup.html.repeater.data.table.PropertyColumn;
import org.apache.wicket.extensions.markup.html.repeater.data.table.export.AbstractExportableColumn;
import org.apache.wicket.extensions.markup.html.repeater.util.SortableDataProvider;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.form.ChoiceRenderer;
import org.apache.wicket.markup.html.form.EnumChoiceRenderer;
import org.apache.wicket.markup.html.link.AbstractLink;
import org.apache.wicket.markup.html.link.BookmarkablePageLink;
import org.apache.wicket.markup.repeater.Item;
import org.apache.wicket.model.AbstractReadOnlyModel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.Model;
import org.apache.wicket.model.PropertyModel;
import org.apache.wicket.request.mapper.parameter.PageParameters;
import scala.runtime.AbstractFunction1;
import se.su.dsv.scipro.components.AjaxCheckBoxMultipleChoice;
import se.su.dsv.scipro.components.DateColumn;
import se.su.dsv.scipro.components.LinkWrapper;
import se.su.dsv.scipro.components.ListAdapterModel;
import se.su.dsv.scipro.components.datatables.MultipleUsersColumn;
import se.su.dsv.scipro.dataproviders.FilteredDataProvider;
import se.su.dsv.scipro.forum.BasicForumService;
import se.su.dsv.scipro.forum.dataobjects.ForumPost;
import se.su.dsv.scipro.profile.UserLinkPanel;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectService;
import se.su.dsv.scipro.project.ProjectStatus;
import se.su.dsv.scipro.project.ProjectTeamMemberRoles;
import se.su.dsv.scipro.reviewing.MyReviewService;
import se.su.dsv.scipro.reviewing.ReviewerInteractionService;
import se.su.dsv.scipro.security.auth.Authorization;
import se.su.dsv.scipro.security.auth.roles.Roles;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.ProjectTypeService;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.util.PageParameterKeys;

import javax.inject.Inject;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import static java.util.Arrays.asList;

@Authorization(authorizedRoles = {Roles.REVIEWER})
public class ReviewerStartPage extends ReviewerPage {
    public static final String MAIN_MENU_LABEL = "Reviewer";

    @Inject
    private MyReviewService myReviewService;
    @Inject
    private ProjectTypeService projectTypeService;
    @Inject
    private ProjectService projectService;
    @Inject
    private ReviewerInteractionService reviewerInteractionService;
    @Inject
    private BasicForumService basicForumService;

    public ReviewerStartPage() {
        addRoughDraftApprovals();
        addInteractWithReviewer();
    }

    private void addRoughDraftApprovals() {
        add(new Label("undecidedRoughDraftApprovals", new AbstractReadOnlyModel<Object>() {
            @Override
            public Object getObject() {
                long undecided = myReviewService.countUndecidedRoughDraft(SciProSession.get().getUser());
                return undecided == 0 ? "" : undecided;
            }
        }));
        add(new RoughDraftApprovalReviewerPanel("RoughDraftApprovals"));
    }

    private void addInteractWithReviewer() {
        final ProjectService.Filter filter = new ProjectService.Filter();
        filter.setFilterSupervisor(false);
        filter.setSupervisor(SciProSession.get().getUser());
        filter.setStatuses(Collections.singletonList(ProjectStatus.ACTIVE));
        filter.setRoles(Collections.singletonList(ProjectTeamMemberRoles.REVIEWER));
        filter.setProjectTypes(projectTypeService.findAllActive());

        final SortableDataProvider<Project, String> provider = new FilteredDataProvider<>(projectService, Model.of(filter));

        DefaultDataTable<Project, String> table = new DefaultDataTable<>("dataPanel", createColumns(), provider, 1000);
        add(table);

        addFilteringOptions(filter, table);

        add(new Label("unreadSupervisorsPosts", new AbstractReadOnlyModel<Object>() {
            @Override
            public Object getObject() {
                List<Project> projects = projectService.findAll(filter);
                int totalNumberOfUnreadPosts = 0;
                for (Project project : projects) {
                    User user = SciProSession.get().getUser();
                    totalNumberOfUnreadPosts += getUnreadPosts(user, project);
                }
                return (totalNumberOfUnreadPosts == 0) ? "" : totalNumberOfUnreadPosts;
            }
        }));
    }

    private int getUnreadPosts(User user, Project project) {
        int unread = 0;
        for (ForumPost post : reviewerInteractionService.getPosts(project)) {
            if (!basicForumService.isRead(user, post)) {
                unread++;
            }
        }
        return unread;
    }

    private void addFilteringOptions(final ProjectService.Filter filter, final Component table) {
        table.setOutputMarkupId(true);
        add(new AjaxCheckBoxMultipleChoice<ProjectStatus>("status",
                new PropertyModel<Collection<ProjectStatus>>(filter, "statuses"),
                asList(ProjectStatus.values()),
                new EnumChoiceRenderer<ProjectStatus>(this)) {
            @Override
            public void onUpdate(AjaxRequestTarget target) {
                target.add(table);
            }
        });
        add(new AjaxCheckBoxMultipleChoice<ProjectType>("type",
                new PropertyModel<Collection<ProjectType>>(filter, "projectTypes"),
                projectTypeService.findBySupervisorProjects(SciProSession.get().getUser()),
                new ChoiceRenderer<ProjectType>("name", "id")) {
            @Override
            public void onUpdate(AjaxRequestTarget target) {
                target.add(table);
            }
        });
    }

    private List<IColumn<Project, String>> createColumns() {
        List<IColumn<Project, String>> columns = new ArrayList<>();
        columns.add(new DateColumn<Project>(Model.of("Started"), "dateCreated"));
        columns.add(new PropertyColumn<Project, String>(Model.of("Type"), "projectType.name"));
        columns.add(new AbstractExportableColumn<Project, String>(Model.of("Title")) {
            @Override
            public void populateItem(Item<ICellPopulator<Project>> cellItem, String componentId, final IModel<Project> rowModel) {
                cellItem.add(LinkWrapper.apply(componentId, new AbstractFunction1<String, AbstractLink>() {
                    @Override
                    public AbstractLink apply(final String id) {
                        PageParameters pp = new PageParameters();
                        pp.add(PageParameterKeys.MAP.get(Project.class), rowModel.getObject().getId());
                        AbstractLink link = new BookmarkablePageLink<>(id, ReviewerInteractionPage.class, pp);
                        link.setBody(new PropertyModel<>(rowModel, "title"));
                        return link;
                    }
                }));
            }

            @Override
            public IModel<String> getDataModel(IModel<Project> rowModel) {
                return new PropertyModel<>(rowModel, "title");
            }
        });
        columns.add(new MultipleUsersColumn<Project, String>(Model.of("Authors")) {
            @Override
            public IModel<? extends List<User>> getUsers(final IModel<Project> rowModel) {
                return new ListAdapterModel<>(new PropertyModel<Iterable<User>>(rowModel, "ProjectParticipants"));
            }
        });
        columns.add(new AbstractExportableColumn<Project, String>(Model.of("Head supervisor")) {
            @Override
            public void populateItem(Item<ICellPopulator<Project>> cellItem, String componentId, IModel<Project> rowModel) {
                cellItem.add(new UserLinkPanel(componentId, new PropertyModel<User>(rowModel, "headSupervisor")));
            }

            @Override
            public IModel<String> getDataModel(IModel<Project> rowModel) {
                return new PropertyModel<>(rowModel, "headSupervisor.fullName");
            }
        });
        columns.add(new AbstractExportableColumn<Project, String>(Model.of("Unread posts")) {
            @Override
            public IModel<Integer> getDataModel(final IModel<Project> rowModel) {
                return new AbstractReadOnlyModel<Integer>() {
                    @Override
                    public Integer getObject() {
                        return getUnreadPosts(SciProSession.get().getUser(), rowModel.getObject());
                    }
                };
            }
        });
        return columns;
    }
}
