package se.su.dsv.scipro.statistics;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.ajax.markup.html.AjaxLink;
import org.apache.wicket.extensions.markup.html.repeater.data.grid.ICellPopulator;
import org.apache.wicket.extensions.markup.html.repeater.data.sort.SortOrder;
import org.apache.wicket.extensions.markup.html.repeater.data.table.IColumn;
import org.apache.wicket.extensions.markup.html.repeater.data.table.PropertyColumn;
import org.apache.wicket.extensions.markup.html.repeater.data.table.export.AbstractExportableColumn;
import org.apache.wicket.extensions.markup.html.repeater.util.SortableDataProvider;
import org.apache.wicket.markup.html.form.ChoiceRenderer;
import org.apache.wicket.markup.html.form.StatelessForm;
import org.apache.wicket.markup.html.panel.FeedbackPanel;
import org.apache.wicket.markup.repeater.Item;
import org.apache.wicket.model.*;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Sort;
import se.su.dsv.scipro.admin.pages.AdminMailPage;
import se.su.dsv.scipro.components.AjaxDropDownChoice;
import se.su.dsv.scipro.components.ExportableDataPanel;
import se.su.dsv.scipro.components.InfoPanel;
import se.su.dsv.scipro.data.DetachableServiceModel;
import se.su.dsv.scipro.dataproviders.PageAdapter;
import se.su.dsv.scipro.match.ApplicationPeriod;
import se.su.dsv.scipro.match.ApplicationPeriodService;
import se.su.dsv.scipro.match.MatchFollowUpService;
import se.su.dsv.scipro.match.TargetResult;
import se.su.dsv.scipro.profile.UserLinkPanel;

import javax.inject.Inject;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

public class AdminUnmetTargetsStatisticsPage extends AbstractAdminStatisticsPage {

    @Inject
    private MatchFollowUpService matchFollowUpService;
    @Inject
    private ApplicationPeriodService applicationPeriodService;

    private final IModel<ApplicationPeriod> applicationPeriod;

    public AdminUnmetTargetsStatisticsPage() {

        final FeedbackPanel feedback = new FeedbackPanel("feedback");
        feedback.setOutputMarkupId(true);
        add(feedback);

        applicationPeriod = new DetachableServiceModel<>(applicationPeriodService);
        final IModel<List<ApplicationPeriod>> applicationPeriodChoices = new LoadableDetachableModel<List<ApplicationPeriod>>() {
            @Override
            protected List<ApplicationPeriod> load() {
                return applicationPeriodService.findAll(new PageRequest(0, Integer.MAX_VALUE, new Sort(Sort.Direction.DESC, "startDate")));
            }
        };
        if (!applicationPeriodChoices.getObject().isEmpty()) {
            applicationPeriod.setObject(applicationPeriodChoices.getObject().get(0));
        }

        add(new StatelessForm<Void>("filter") {
            {
                add(new AjaxDropDownChoice<ApplicationPeriod>("applicationPeriod", applicationPeriod, applicationPeriodChoices, new ChoiceRenderer<>("name", "id")) {
                    {
                        setNullValid(false);
                        setRequired(true);
                    }

                    @Override
                    public void onNewSelection(final AjaxRequestTarget target, final ApplicationPeriod objectSelected) {
                        target.add(getPage().get("table"), getPage().get("mailLink"));
                    }
                });

                add(new InfoPanel("info", new ResourceModel("info")));
            }
        });

        UnmetTargetsProvider unmetTargetsProvider = new UnmetTargetsProvider();
        unmetTargetsProvider.setSort("user.lastName", SortOrder.ASCENDING);

        add(new AjaxLink<Void>("mailLink") {
            {
                setOutputMarkupPlaceholderTag(true);
            }

            @Override
            public void onClick(AjaxRequestTarget target) {
                setResponsePage(
                        new AdminMailPage(matchFollowUpService.findSupervisorsWithUnmetTargets(applicationPeriod.getObject()),
                                "Add ideas in SciPro",
                                "You don't have enough ideas to meet your targets for application period " + applicationPeriod.getObject().getName()));
            }
            @Override
            protected void onConfigure() {
                setVisibilityAllowed(applicationPeriod.getObject()!=null);
            }
        });

        add(new ExportableDataPanel<TargetResult, String>("table", columns(), unmetTargetsProvider) {
            {
                setOutputMarkupPlaceholderTag(true);
            }

            @Override
            protected void onConfigure() {
                setVisible(applicationPeriod.getObject() != null);
            }
        });
    }

    private List<? extends IColumn<TargetResult, String>> columns() {
        return Arrays.asList(
                new AbstractExportableColumn<TargetResult, String>(new StringResourceModel("th.supervisor", this), "user.lastName") {
                    @Override
                    public void populateItem(final Item<ICellPopulator<TargetResult>> cellItem, final String componentId, final IModel<TargetResult> rowModel) {
                        cellItem.add(new UserLinkPanel(componentId, new PropertyModel<>(rowModel, "target.user")));
                    }

                    @Override
                    public IModel<String> getDataModel(final IModel<TargetResult> rowModel) {
                        return new PropertyModel<>(rowModel, "target.user.fullName");
                    }
                },
                new PropertyColumn<>(new StringResourceModel("th.projectType", this), "projectType.name", "target.projectType.name"),
                new PropertyColumn<>(new StringResourceModel("th.target", this), "target", "target.target"),
                new PropertyColumn<>(new StringResourceModel("th.submittedAndTaken", this), "submittedAndTaken"),
                new AbstractExportableColumn<TargetResult, String>(Model.of("Left")) {
                    @Override
                    public IModel<String> getDataModel(final IModel<TargetResult> rowModel) {
                        final TargetResult targetResult = rowModel.getObject();
                        return Model.of(String.valueOf(targetResult.getTarget().getTarget() - targetResult.getSubmittedAndTaken()));
                    }
                }
        );
    }

    private class UnmetTargetsProvider extends SortableDataProvider<TargetResult, String> {
        @Override
        public Iterator<? extends TargetResult> iterator(final long first, final long count) {
            return matchFollowUpService.findUnmetTargets(applicationPeriod.getObject(), new PageAdapter(first, count, getSort())).iterator();
        }

        @Override
        public long size() {
            return matchFollowUpService.countUnmetTargets(applicationPeriod.getObject());
        }

        @Override
        public IModel<TargetResult> model(final TargetResult object) {
            return Model.of(object);
        }
    }
}
